# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""VariantCall utilities."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from nucleus.protos import struct_pb2
from nucleus.protos import variants_pb2
from nucleus.util import struct_utils
from nucleus.util import vcf_constants

# Special-cased FORMAT fields that are first-class fields in the VariantCall.
_GL = 'GL'
_GT = 'GT'


def set_format(variant_call, field_name, value, vcf_object=None):
  """Sets a field of the info map of the `VariantCall` to the given value(s).

  `variant_call.info` is analogous to the FORMAT field of a VCF call.

  Example usage:
  with vcf.VcfReader('/path/to/my.vcf') as vcf_reader:
    for variant in vcf_reader:
      first_call = variant.calls[0]
      # Type can be inferred for reserved VCF fields.
      set_format(first_call, 'AD', 25)
      # Specify the reader explicitly for unknown fields.
      set_format(first_call, 'MYFIELD', 30, vcf_reader)

  Args:
    variant_call: VariantCall proto. The VariantCall to modify.
    field_name: str. The name of the field to set.
    value: A single value or list of values to update the VariantCall with.
      The type of the value is determined by the `vcf_object` if one is given,
      otherwise is looked up based on the reserved FORMAT fields in the VCF
      specification.
    vcf_object: (Optional) A VcfReader or VcfWriter object. If not None, the
      type of the field is inferred from the associated VcfReader or VcfWriter
      based on its name. Otherwise, the type is inferred if it is a reserved
      field.
  """
  if field_name == _GL:
    set_gl(variant_call, value)
    return
  if field_name == _GT:
    set_gt(variant_call, value)
    return

  if vcf_object is None:
    set_field_fn = vcf_constants.reserved_format_field_set_fn(field_name)
  else:
    set_field_fn = vcf_object.field_access_cache.format_field_set_fn(field_name)
  set_field_fn(variant_call.info, field_name, value)


def get_format(variant_call, field_name, vcf_object=None):
  """Returns the value of the `field_name` FORMAT field.

  The `vcf_object` is used to determine the type of the resulting value. If it
  is a single value or a Flag, that single value will be returned. Otherwise,
  the list of values is returned.

  Args:
    variant_call: VariantCall proto. The VariantCall of interest.
    field_name: str. The name of the field to retrieve values from.
    vcf_object: (Optional) A VcfReader or VcfWriter object. If not None, the
      type of the field is inferred from the associated VcfReader or VcfWriter
      based on its name. Otherwise, the type is inferred if it is a reserved
      field.
  """
  if field_name == _GL:
    return get_gl(variant_call)
  if field_name == _GT:
    return get_gt(variant_call)

  if vcf_object is None:
    get_field_fn = vcf_constants.reserved_format_field_get_fn(field_name)
  else:
    get_field_fn = vcf_object.field_access_cache.format_field_get_fn(field_name)
  return get_field_fn(variant_call.info, field_name)


# The following functions are convenience methods for getting/setting some
# reserved FORMAT fields of a VariantCall as well as some non-reserved FORMAT
# fields used by DeepVariant. Note that these functions will use the types of
# each field as defined by the VCF 4.3 specification, mirrored in
# vcf_constants.py, so if you have redefined any of these fields to have
# different types these functions will not do what you want.
def set_ad(variant_call, ad):
  """Sets the allele depth of the VariantCall."""
  set_format(variant_call, 'AD', ad)


def get_ad(variant_call):
  """Gets the allele depth of the VariantCall."""
  return get_format(variant_call, 'AD')


def set_gl(variant_call, gl):
  """Sets the genotype likelihoods of the VariantCall.

  Args:
    variant_call: VariantCall proto. The VariantCall to modify.
    gl: list(float). The list of genotype likelihoods for the VariantCall.
  """
  # Note: genotype_likelihood is extracted to a first-class field within
  # VariantCall. Consequently, we just set its value directly here.
  variant_call.genotype_likelihood[:] = gl


def get_gl(variant_call):
  """Returns the genotype likelihoods of the VariantCall.

  Args:
    variant_call: VariantCall proto. The VariantCall for which to return GLs.

  Returns:
    A list of floats representing the genotype likelihoods of this call.
  """
  return variant_call.genotype_likelihood


def set_gt(variant_call, gt):
  """Sets the genotypes of the VariantCall.

  Args:
    variant_call: VariantCall proto. The VariantCall to modify.
    gt: list(int). The list of genotypes for the VariantCall.
  """
  # Note: genotype is extracted to a first-class field within
  # VariantCall. Consequently, we just set its value directly here.
  variant_call.genotype[:] = gt


def get_gt(variant_call):
  """Returns the genotypes of the VariantCall.

  Args:
    variant_call: VariantCall proto. The VariantCall for which to return GTs.

  Returns:
    A list of ints representing the genotype indices of this call.
  """
  return variant_call.genotype


def set_gq(variant_call, gq):
  """Sets the genotype quality of the VariantCall."""
  set_format(variant_call, 'GQ', gq)


def get_gq(variant_call):
  """Gets the genotype quality of the VariantCall."""
  return get_format(variant_call, 'GQ')


def set_min_dp(variant_call, min_dp):
  """Sets the 'MIN_DP' field of the VariantCall."""
  struct_utils.set_int_field(variant_call.info, 'MIN_DP', min_dp)


def get_min_dp(variant_call):
  """Gets the 'MIN_DP' field of the VariantCall."""
  return struct_utils.get_int_field(
      variant_call.info, 'MIN_DP', is_single_field=True)


def has_genotypes(variant_call):
  """Returns True iff the VariantCall has one or more called genotypes.

  Args:
    variant_call: VariantCall proto. The VariantCall to evaluate.

  Returns:
    True if the VariantCall has one or more called genotypes, False otherwise.
  """
  return any(gt >= 0 for gt in variant_call.genotype)


def has_full_genotypes(variant_call):
  """Returns True iff the VariantCall has only known genotypes.

  Args:
    variant_call: VariantCall proto. The VariantCall to evaluate.

  Returns:
    True if all `genotype` fields are known genotypes.
  """
  return all(gt >= 0 for gt in variant_call.genotype)


def ploidy(variant_call):
  """Returns the ploidy of the VariantCall.

  Args:
    variant_call: VariantCall proto. The VariantCall to evaluate.

  Returns:
    The ploidy of the call (a non-negative integer).
  """
  # Unknown genotypes are represented as -1 in VariantCall protos. When
  # a VCF is parsed that contains multiple ploidies in different samples,
  # a separate padding value of -2**30 - 1 is inserted into the calls.
  return sum(gt >= -1 for gt in variant_call.genotype)


def has_variation(variant_call):
  """Returns True if and only if the call has a non-reference genotype.

  Args:
    variant_call: VariantCall proto. The VariantCall to evaluate.

  Returns:
    True if and only if the call has a non-reference genotype.
  """
  return any(gt > 0 for gt in variant_call.genotype)


def is_heterozygous(variant_call):
  """Returns True if and only if the call is heterozygous.

  Args:
    variant_call: VariantCall proto. The VariantCall to evaluate.

  Returns:
    True if and only if the call is heterozygous.
  """
  return len({gt for gt in variant_call.genotype if gt >= 0}) >= 2



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes that provide the interface for writing genomics data.

`GenomicsWriter` defines the core API supported by writers, and is subclassed
directly or indirectly (via `DispatchingGenomicsWriter`) for all concrete
implementations.

`TFRecordWriter` is an implementation of the `GenomicsWriter` API for reading
`TFRecord` files. This is usable for all data types when writing data as
serialized protocol buffers.

`DispatchingGenomicsWriter` is an abstract class defined for convenience on top
of `GenomicsWriter` that supports writing to either the native file format or to
`TFRecord` files of the corresponding protocol buffer used to encode data of
that file type. The output format chosen is dependent upon the filename to which
the data are being written.

Concrete implementations for individual file types (e.g. BED, SAM, VCF, etc.)
reside in type-specific modules in this package. A general example of the write
functionality is shown below.

```python
# options is a writer-specific set of options.
options = ...

# records is an iterable of protocol buffers of the specific data type.
records = ...

with GenomicsWriterSubClass(output_path, options) as writer:
  for proto in records:
    writer.write(proto)
```
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import errno

from absl import logging

from nucleus.io.python import tfrecord_writer


class GenomicsWriter(object):
  """Abstract base class for writing genomics data.

  A GenomicsWriter only has one method, write, which writes a single
  protocol buffer to a file.
  """

  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def write(self, proto):
    """Writes proto to the file.

    Args:
      proto:  A protocol buffer.
    """

  def __enter__(self):
    """Enter a `with` block."""
    return self

  @abc.abstractmethod
  def __exit__(self, unused_type, unused_value, unused_traceback):
    """Exit a `with` block.  Typically, this will close the file."""


class TFRecordWriter(GenomicsWriter):
  """A GenomicsWriter that writes to a TFRecord file.

  Example usage:
    writer = TFRecordWriter('/tmp/my_output.tfrecord.gz')
    for record in records:
      writer.write(record)

  Note that TFRecord files do not need to be wrapped in a "with" block.
  """

  def __init__(self, output_path, header=None, compression_type=None):
    """Initializer.

    Args:
      output_path: str. The output path to which the records are written.
      header: An optional header for the particular data type. This can be
        useful for file types that have logical headers where some operations
        depend on that header information (e.g. VCF using its headers to
        determine type information of annotation fields).
      compression_type:  Either 'ZLIB', 'GZIP', '' (uncompressed), or
        None.  If None, __init__ will guess the compression type based on
        the input_path's suffix.

    Raises:
      IOError:  if there was any problem opening output_path for writing.
    """
    super(TFRecordWriter, self).__init__()
    self.header = header

    if compression_type is None:
      compression_type = 'GZIP' if output_path.endswith('.gz') else ''

    self._writer = tfrecord_writer.TFRecordWriter.from_file(
        output_path, compression_type)
    if self._writer is None:
      raise IOError(errno.EIO, 'Error opening %s for writing' % output_path)

  def write(self, proto):
    """Writes the proto to the TFRecord file."""
    self._writer.write(proto.SerializeToString())

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._writer.close()


class DispatchingGenomicsWriter(GenomicsWriter):
  """A GenomicsWriter that dispatches based on the file extension.

  If '.tfrecord' is present in the filename, a TFRecordWriter is used.
  Otherwise, a native writer is.

  Sub-classes of DispatchingGenomicsWriter must define a _native_writer()
  method.
  """

  def __init__(self, output_path, **kwargs):
    """Initializer.

    Args:
      output_path: str. The output path to which the records are written.
      **kwargs: k=v named args. Keyword arguments used to instantiate the native
        writer, if applicable.
    """
    super(DispatchingGenomicsWriter, self).__init__()
    self.header = kwargs.get('header', None)

    if '.tfrecord' in output_path:
      self._writer = TFRecordWriter(output_path, header=self.header)
    else:
      self._writer = self._native_writer(output_path, **kwargs)
    logging.info('Writing %s with %s',
                 output_path, self._writer.__class__.__name__)
    self._post_init_hook()

  @abc.abstractmethod
  def _native_writer(self, output_path, **kwargs):
    """Returns a GenomicsWriter for writing the records `natively`.

    Args:
      output_path: The path to write the records to.
      **kwargs:  Zero or more keyword arguments.

    Returns:
      A GenomicsWriter.
    """

  def write(self, proto):
    self._writer.write(proto)

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._writer.__exit__(exit_type, exit_value, exit_traceback)

  def _post_init_hook(self):
    """Hook for subclasses to run code at the end of __init__."""



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A universal converter program for nucleus-supported genomics file formats.

Invoked with a single argument, this program will open a genomics data file and
iterate over its contents, doing no writing.  This is a good benchmark for I/O
and reader processing speed.

Invoked with two arguments, the program will open the first file, read its
records, and write them, one at a time, to the second file.  The filetypes for
the first and second filename must be compatible ways of encoding the same
nucleus genomics record type (for example, `infile.gff` and
`outfile.gff.tfrecord.gz` are compatible, but `infile.gff` and `outfile.bam` are
not.

Note: at present we have no convention for encoding a file *header* in
tfrecords, so conversion is not possible from tfrecord to any native file format
for which a header is compulsory.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import re
import sys
import time

from absl import app
from absl import logging

from nucleus.io import bed
from nucleus.io import fastq
from nucleus.io import gff
from nucleus.io import sam
from nucleus.io import vcf


def _is_native_file(filename):
  """Returns true if filename is a native (non-tfrecord) genomics data file."""
  return not re.match(r".*\.tfrecord(\.gz)?", filename)


def _filename_pattern(ext):
  """Returns an re matching native or tfrecord files of format `ext`."""
  return r".*\.{}(\.tfrecord)?(\.gz)?".format(ext)


_FileType = collections.namedtuple(
    "_FileType", ("reader_class", "writer_class", "has_header"))

_FILETYPE_LOOKUP = {
    _filename_pattern("bed"):
        _FileType(bed.BedReader, bed.BedWriter, False),
    _filename_pattern("(fastq|fq)"):
        _FileType(fastq.FastqReader, fastq.FastqWriter, False),
    _filename_pattern("gff"):
        _FileType(gff.GffReader, gff.GffWriter, True),
    _filename_pattern("(bam|sam)"):
        _FileType(sam.SamReader, sam.SamWriter, True),
    _filename_pattern("vcf"):
        _FileType(vcf.VcfReader, vcf.VcfWriter, True),
}


def _lookup_filetype(filename):
  for pattern in _FILETYPE_LOOKUP:
    if re.match(pattern, filename):
      return _FILETYPE_LOOKUP[pattern]
  raise ConversionError("Unrecognized extension!")

LOG_EVERY = 100000


class ConversionError(Exception):
  """An exception used to signal file conversion error."""
  pass


class NullWriter(object):
  """A writer class whose .write() method is a no-op.

  This allows us to create and use a writer object where one is required by
  context but we do not wish to write to any file.
  """

  def __init__(self, unused_filename, header=None):
    pass

  def write(self, unused_record):
    pass

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    pass


def _reader_writer_classes(in_filename, out_filename):
  """Returns reader, writer classes for filenames, if conversion is possible.

  Args:
    in_filename: filename of a genomics data file to use as input.
    out_filename: filename of a genomics data file to use as output, or None,
      if no output should be written.

  Raises:
    ConversionError: if in_filename is not convertible to out_filename.
  """
  in_filetype = _lookup_filetype(in_filename)
  out_filetype = _lookup_filetype(out_filename) if out_filename else None

  if out_filetype:
    if in_filetype != out_filetype:
      raise ConversionError(
          "Input and output filetypes specified are incompatible.")
    input_has_header = in_filetype.has_header and _is_native_file(in_filename)
    output_requires_header = (
        out_filetype.has_header and _is_native_file(out_filename))
    if output_requires_header and not input_has_header:
      raise ConversionError(
          "Input file does not have a header, which is needed to construct "
          "output file")
    writer_class = out_filetype.writer_class

  else:
    writer_class = NullWriter

  return in_filetype.reader_class, writer_class


def convert(in_filename, out_filename):
  """Converts a recognized genomics file `in_filename` to `out_filename`.

  Args:
    in_filename: str; filename of a genomics data file to use as input.
    out_filename: str; filename of a genomics data file to use as output, or
      None, if no output should be written.

  Raises:
    ConversionError, if the conversion could not be executed.
  """
  reader_class, writer_class = _reader_writer_classes(in_filename, out_filename)
  reader = reader_class(in_filename)

  with reader_class(in_filename) as reader:
    with writer_class(out_filename, header=reader.header) as writer:
      start = time.time()
      i = 0
      for record in reader:
        i += 1
        writer.write(record)
        logging.log_every_n(logging.INFO, "Progress: %d records", LOG_EVERY, i)
      elapsed = time.time() - start
      logging.info("Done, processed %d records in %0.2f seconds.", i, elapsed)


def main(argv):
  if len(argv) not in (2, 3):
    print("Usage: %s <input_filename> [<output_filename>]" % argv[0])
    sys.exit(1)

  input_filename = argv[1]
  output_filename = None if len(argv) == 2 else argv[2]

  try:
    convert(input_filename, output_filename)
  except ConversionError as e:
    print("Could not execute conversion:", e)
    sys.exit(1)


if __name__ == "__main__":
  app.run(main)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for working with alignment CIGAR operations.

The CIGAR format is defined within the SAM spec, available at
https://samtools.github.io/hts-specs/SAMv1.pdf

This module provides utility functions for interacting with the parsed
representations of CIGAR strings.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import six
from nucleus.protos import cigar_pb2

# A frozenset of all CigarUnit.Operation enum values that advance the alignment
# with respect to the reference genome and read, respectively.
REF_ADVANCING_OPS = frozenset([
    cigar_pb2.CigarUnit.ALIGNMENT_MATCH,
    cigar_pb2.CigarUnit.SEQUENCE_MATCH,
    cigar_pb2.CigarUnit.DELETE,
    cigar_pb2.CigarUnit.SKIP,
    cigar_pb2.CigarUnit.SEQUENCE_MISMATCH
])
READ_ADVANCING_OPS = frozenset([
    cigar_pb2.CigarUnit.ALIGNMENT_MATCH,
    cigar_pb2.CigarUnit.SEQUENCE_MATCH,
    cigar_pb2.CigarUnit.INSERT,
    cigar_pb2.CigarUnit.CLIP_SOFT,
    cigar_pb2.CigarUnit.SEQUENCE_MISMATCH
])

# A map from CigarUnit.Operation (e.g., CigarUnit.ALIGNMENT_MATCH) enum values
# to their corresponding single character cigar codes (e.g., 'M').
CIGAR_OPS_TO_CHAR = {
    cigar_pb2.CigarUnit.ALIGNMENT_MATCH: 'M',
    cigar_pb2.CigarUnit.INSERT: 'I',
    cigar_pb2.CigarUnit.DELETE: 'D',
    cigar_pb2.CigarUnit.SKIP: 'N',
    cigar_pb2.CigarUnit.CLIP_SOFT: 'S',
    cigar_pb2.CigarUnit.CLIP_HARD: 'H',
    cigar_pb2.CigarUnit.PAD: 'P',
    cigar_pb2.CigarUnit.SEQUENCE_MATCH: '=',
    cigar_pb2.CigarUnit.SEQUENCE_MISMATCH: 'X',
}

# A map from single character cigar codes (e.g., 'M') to their corresponding
# CigarUnit.Operation (e.g., CigarUnit.ALIGNMENT_MATCH) enum values.
CHAR_TO_CIGAR_OPS = {v: k for k, v in CIGAR_OPS_TO_CHAR.items()}

# All of the CigarUnit.Operation values in a frozen set.
ALL_CIGAR_OPS = frozenset(CIGAR_OPS_TO_CHAR.keys())

# Regular expression that matches only valid full cigar strings.
VALID_CIGAR_RE = re.compile(
    r'^(\d+[' + ''.join(CHAR_TO_CIGAR_OPS.keys()) + '])+$')

# Regular expression that matches a single len/op cigar element. The match is
# grouped, so CIGAR_STR_SPLITTER_RE.finditer(cigar_str) returns grouped units
# of the cigar string in order.
CIGAR_STR_SPLITTER_RE = re.compile(
    r'(\d+[' + ''.join(CHAR_TO_CIGAR_OPS.keys()) + '])')


def format_cigar_units(cigar_units):
  """Returns the string version of an iterable of CigarUnit protos.

  Args:
    cigar_units: iterable[CigarUnit] protos.

  Returns:
    A string representation of the CigarUnit protos that conforms to the
    CIGAR string specification.
  """
  return ''.join(
      str(unit.operation_length) + CIGAR_OPS_TO_CHAR[unit.operation]
      for unit in cigar_units)


def parse_cigar_string(cigar_str):
  """Parse a cigar string into a list of cigar units.

  For example, if cigar_str is 150M2S, this function will return:

  [
    CigarUnit(operation=ALIGNMENT_MATCH, operation_length=150),
    CigarUnit(operation=SOFT_CLIP, operation_length=2)
  ]

  Args:
    cigar_str: str containing a valid cigar.

  Returns:
    list[cigar_pb2.CigarUnit].

  Raises:
    ValueError: If cigar_str isn't a well-formed CIGAR.
  """
  if not cigar_str:
    raise ValueError('cigar_str cannot be empty')
  if not VALID_CIGAR_RE.match(cigar_str):
    raise ValueError('Malformed CIGAR string {}'.format(cigar_str))
  parts = CIGAR_STR_SPLITTER_RE.finditer(cigar_str)
  return [to_cigar_unit(part.group(1)) for part in parts]


def alignment_length(cigar_units):
  """Computes the span in basepairs of the cigar units.

  Args:
    cigar_units: iterable[CigarUnit] whose alignment length we want to compute.

  Returns:
    The number of basepairs spanned by the cigar_units.
  """
  return sum(unit.operation_length
             for unit in cigar_units
             if unit.operation in REF_ADVANCING_OPS)


def to_cigar_unit(source):
  """Creates a cigar_pb2 CigarUnit from source.

  This function attempts to convert source into a CigarUnit protobuf. If
  source is a string, it must be a single CIGAR string specification like
  '12M'. If source is a tuple or a list, must have exactly two elements
  (operation_length, opstr). operation_length can be a string or int, and must
  be >= 1. opstr should be a single character CIGAR specification (e.g., 'M').
  If source is already a CigarUnit, it is just passed through unmodified.

  Args:
    source: many types allowed. The object we want to convert to a CigarUnit
      proto.

  Returns:
    CigarUnit proto with operation_length and operation set to values from
      source.

  Raises:
    ValueError: if source cannot be converted or is malformed.
  """
  try:
    if isinstance(source, cigar_pb2.CigarUnit):
      return source
    elif isinstance(source, six.string_types):
      l, op = source[:-1], source[-1]
    elif isinstance(source, (tuple, list)):
      l, op = source
    else:
      raise ValueError('Unexpected source', source)

    if isinstance(op, six.string_types):
      op = CHAR_TO_CIGAR_OPS[op]
    l = int(l)
    if l < 1:
      raise ValueError('Length must be >= 1', l)
    return cigar_pb2.CigarUnit(operation=op, operation_length=int(l))
  except (KeyError, IndexError):
    raise ValueError('Failed to convert {} into a CigarUnit'.format(source))


def to_cigar_units(source):
  """Converts object to a list of CigarUnit.

  This function attempts to convert source into a list of CigarUnit protos.
  If source is a string, we assume it is a CIGAR string and call
  parse_cigar_string on it, returning the result. It not, we assume it's an
  iterable containing element to be converted with to_cigar_unit(). The
  resulting list of converted elements is returned.

  Args:
    source: str or iterable to convert to CigarUnit protos.

  Returns:
    list[CigarUnit].
  """
  if isinstance(source, six.string_types):
    return parse_cigar_string(source)
  else:
    return [to_cigar_unit(singleton) for singleton in source]



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Given new genotype priors, update the variant calls in a VCF or gVCF file."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math

from absl import app
from absl import flags

from nucleus.io import vcf
from nucleus.util import variant_utils
from nucleus.util import variantcall_utils

FLAGS = flags.FLAGS

flags.DEFINE_string(
    'input_vcf', None,
    'Filename of the VCF file to read.')
flags.DEFINE_string(
    'output_vcf', None,
    'Filename of the VCF file to write to.')
flags.DEFINE_list(
    'genotype_priors', [],
    'List of genotype priors.  The length of this list should be the same '
    'as the length of the PL or GL field in the input_vcf.')
flags.DEFINE_float(
    'posterior_margin', 0.0,
    'If the difference between the top two posteriors is less than this, '
    'make no call (i.e., call it as "./.").')


def recall_variant(log_priors, variant):
  """Update the genotype calls in variant given the new genotype priors."""
  for call in variant.calls:
    if len(log_priors) != len(call.genotype_likelihood):
      continue

    posteriors = [x + y for x, y in zip(log_priors, call.genotype_likelihood)]
    sorted_posts = sorted(posteriors)
    highest_post = sorted_posts[-1]

    margin = 0
    if FLAGS.posterior_margin > 0.0 and len(log_priors) > 1:
      margin = math.pow(10.0, highest_post) - math.pow(10.0, sorted_posts[-2])

    ploidy = variantcall_utils.ploidy(call)
    if margin < FLAGS.posterior_margin:
      call.genotype[:] = [-1] * ploidy
    else:
      best_genotype = posteriors.index(highest_post)
      call.genotype[:] = (
          variant_utils.allele_indices_for_genotype_likelihood_index(
              best_genotype, ploidy=ploidy))


def main(argv):
  del argv

  priors = map(float, FLAGS.genotype_priors)
  sump = sum(priors)
  log_priors = [math.log10(x / sump) for x in priors]

  with vcf.VcfReader(FLAGS.input_vcf) as reader:
    with vcf.VcfWriter(FLAGS.output_vcf, header=reader.header) as writer:
      for variant in reader:
        recall_variant(log_priors, variant)
        # TODO(thomaswc): Also update the variant's quality.
        writer.write(variant)


if __name__ == '__main__':
  flags.mark_flags_as_required(['input_vcf', 'output_vcf'])
  app.run(main)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=line-too-long
"""Classes for reading and writing SAM and BAM files.

The SAM/BAM/CRAM formats are described at
https://samtools.github.io/hts-specs/SAMv1.pdf
https://samtools.github.io/hts-specs/CRAMv3.pdf

API for reading:

```python
from nucleus.io import sam

with sam.SamReader(input_path) as reader:
  for read in reader:
    print(read)
```

where `read` is a `nucleus.genomics.v1.Read` protocol buffer. input_path will
dynamically decode the underlying records depending the file extension, with
`.sam` for SAM files, `.bam` for BAM files, and `.cram` for CRAM files. It will
also search for an appropriate index file to use to enable calls to the
`query()` method.

API for writing SAM/BAM:

```python
from nucleus.io import sam

# reads is an iterable of nucleus.genomics.v1.Read protocol buffers.
reads = ...

with sam.SamWriter(output_path, header=header) as writer:
  for read in reads:
    writer.write(read)
```

API for writing CRAM:

```python
# ref_path is required for writing CRAM files. If embed_ref, the output CRAM
# file will embed reference sequences.
with sam.SamWriter(output_path, header=header, ref_path=ref_path,
                   embed_ref=embed_ref) as writer:
  for read in reads:
    writer.write(read)
```

For both reading and writing, if the path provided to the constructor contains
'.tfrecord' as an extension, a `TFRecord` file is assumed and attempted to be
read or written. Otherwise, the filename is treated as a true SAM/BAM/CRAM file.

For `TFRecord` files, ending in a '.gz' suffix causes the file to be treated as
compressed with gzip.

Notes on using CRAM with SamReader
--------------------------------

Nucleus supports reading from CRAM files using the same API as for SAM/BAM:

```python
from nucleus.io import sam

with sam.SamReader("/path/to/sample.cram") as reader:
  for read in reader:
    print(read)
```

There is one type of CRAM file, though, that has a slightly more complicated
API. If the CRAM file uses read sequence compression with an external reference
file, and this reference file is no longer accessible in the location specified
by the CRAM file's "UR" tag and cannot be found in the local genome cache, its
location must be passed to SamReader via the ref_path parameter:

```python
from nucleus.io import sam

cram_path = "/path/to/sample.cram"
ref_path = "/path/to/genome.fasta"
with sam.SamReader(cram_path, ref_path=ref_path) as reader:
  for read in reader:
    print(read)
```

Unfortunately, htslib is unable to load the ref_path from anything other than a
POSIX filesystem. (htslib plugin filesystems like S3 or GCS buckets won't work).
For that reason, we don't recommend the use of CRAM files with external
reference files, but instead suggest using read sequence compression with
embedded reference data. (This has a minor impact on file size, but
significantly improves file access simplicity and safety.)

For more information about CRAM, see:
* The `samtools` documentation at http://www.htslib.org/doc/samtools.html
* The "Global Options" section of the samtools docs at http://www.htslib.org/doc/samtools.html#GLOBAL_OPTIONS
* How reference sequences are encoded in CRAM at http://www.htslib.org/doc/samtools.html#REFERENCE_SEQUENCES
* Finally, benchmarking of different CRAM options http://www.htslib.org/benchmarks/CRAM.html
"""
# pylint: enable=line-too-long

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from nucleus.io import genomics_reader
from nucleus.io import genomics_writer
from nucleus.io.python import sam_reader
from nucleus.io.python import sam_writer
from nucleus.protos import reads_pb2
from nucleus.util import ranges
from nucleus.util import utils


class NativeSamReader(genomics_reader.GenomicsReader):
  """Class for reading from native SAM/BAM/CRAM files.

  Most users will want to use SamReader instead, because it dynamically
  dispatches between reading native SAM/BAM/CRAM files and TFRecord files based
  on the filename's extensions.
  """

  def __init__(self,
               input_path,
               ref_path=None,
               read_requirements=None,
               parse_aux_fields=False,
               hts_block_size=None,
               downsample_fraction=None,
               random_seed=None,
               use_original_base_quality_scores=False):
    """Initializes a NativeSamReader.

    Args:
      input_path: str. A path to a resource containing SAM/BAM/CRAM records.
        Currently supports SAM text format, BAM binary format, and CRAM.
      ref_path: optional str or None. Only used for CRAM decoding, and only
        necessary if the UR encoded path in the CRAM itself needs to be
        overridden. If provided, we will tell the CRAM decoder to use this FASTA
        for the reference sequence.
      read_requirements: optional ReadRequirement proto. If not None, this proto
        is used to control which reads are filtered out by the reader before
        they are passed to the client.
      parse_aux_fields: optional bool, defaulting to False. If False, we do not
        parse the auxiliary fields of the SAM/BAM/CRAM records (see SAM spec for
        details). Parsing the aux fields is unnecessary for many applications,
        and adds a significant parsing cost to access. If you need these aux
        fields, set parse_aux_fields to True and these fields will be parsed and
        populate the appropriate Read proto fields (e.g., read.info).
      hts_block_size: int or None. If specified, this configures the block size
        of the underlying htslib file object. Larger values (e.g. 1M) may be
        beneficial for reading remote files. If None, the reader uses the
        default htslib block size.
      downsample_fraction: float in the interval [0.0, 1.0] or None. If
        specified as a positive float, the reader will only keep each read with
        probability downsample_fraction, randomly. If None or zero, all reads
        are kept.
      random_seed: None or int. The random seed to use with this sam reader, if
        needed. If None, a fixed random value will be assigned.
      use_original_base_quality_scores: optional bool, defaulting to False. If
        True, quality scores are read from OQ tag.

    Raises:
      ValueError: If downsample_fraction is not None and not in the interval
        (0.0, 1.0].
      ImportError: If someone tries to load a tfbam file.
    """
    if input_path.endswith('.tfbam'):
      # Delayed loading of tfbam_lib.
      try:
        from tfbam_lib import tfbam_reader  # pylint: disable=g-import-not-at-top
        self._reader = tfbam_reader.make_sam_reader(
            input_path,
            read_requirements=read_requirements,
            unused_block_size=hts_block_size,
            downsample_fraction=downsample_fraction,
            random_seed=random_seed)
      except ImportError:
        raise ImportError(
            'tfbam_lib module not found, cannot read .tfbam files.')
    else:
      aux_field_handling = reads_pb2.SamReaderOptions.SKIP_AUX_FIELDS
      if parse_aux_fields:
        aux_field_handling = reads_pb2.SamReaderOptions.PARSE_ALL_AUX_FIELDS

      # We make 0 be a valid value that means "keep all reads" so that proto
      # defaults (=0) do not omit all reads.
      if downsample_fraction is not None and downsample_fraction != 0:
        if not 0.0 < downsample_fraction <= 1.0:
          raise ValueError(
              'downsample_fraction must be in the interval (0.0, 1.0]',
              downsample_fraction)

      if random_seed is None:
        # Fixed random seed produced with 'od -vAn -N4 -tu4 < /dev/urandom'.
        random_seed = 2928130004

      self._reader = sam_reader.SamReader.from_file(
          input_path.encode('utf8'),
          ref_path.encode('utf8') if ref_path is not None else '',
          reads_pb2.SamReaderOptions(
              read_requirements=read_requirements,
              aux_field_handling=aux_field_handling,
              hts_block_size=(hts_block_size or 0),
              downsample_fraction=downsample_fraction,
              random_seed=random_seed,
              use_original_base_quality_scores=use_original_base_quality_scores)
      )

      self.header = self._reader.header

    super(NativeSamReader, self).__init__()

  def iterate(self):
    """Returns an iterable of Read protos in the file."""
    return self._reader.iterate()

  def query(self, region):
    """Returns an iterator for going through the reads in the region."""
    return self._reader.query(region)

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._reader.__exit__(exit_type, exit_value, exit_traceback)


class SamReader(genomics_reader.DispatchingGenomicsReader):
  """Class for reading Read protos from SAM/BAM/CRAM or TFRecord files."""

  def _native_reader(self, input_path, **kwargs):
    return NativeSamReader(input_path, **kwargs)

  def _record_proto(self):
    return reads_pb2.Read


class NativeSamWriter(genomics_writer.GenomicsWriter):
  """Class for writing to native SAM/BAM/CRAM files.

  Most users will want SamWriter, which will write to either native SAM/BAM/CRAM
  files or TFRecords files, based on the output filename's extensions.
  """

  def __init__(self, output_path, header, ref_path=None, embed_ref=False):
    """Initializer for NativeSamWriter.

    Args:
      output_path: str. A path where we'll write our SAM/BAM/CRAM file.
      ref_path: str. Path to the reference file. Required for CRAM file.
      embed_ref: bool. Whether to embed the reference sequences in CRAM file.
        Default is False.
      header: A nucleus.SamHeader proto.  The header is used both for writing
        the header, and to control the sorting applied to the rest of the file.
    """
    super(NativeSamWriter, self).__init__()
    self._writer = sam_writer.SamWriter.to_file(
        output_path,
        ref_path.encode('utf8') if ref_path is not None else '', embed_ref,
        header)

  def write(self, proto):
    self._writer.write(proto)

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._writer.__exit__(exit_type, exit_value, exit_traceback)


class SamWriter(genomics_writer.DispatchingGenomicsWriter):
  """Class for writing Read protos to SAM or TFRecord files."""

  def _native_writer(self, output_path, **kwargs):
    return NativeSamWriter(output_path, **kwargs)


class InMemorySamReader(object):
  """Python interface class for in-memory SAM/BAM/CRAM reader.

  Attributes:
    reads: list[nucleus.genomics.v1.Read]. The list of in-memory reads.
    is_sorted: bool, True if reads are sorted.
  """

  def __init__(self, reads, is_sorted=False):
    self.replace_reads(reads, is_sorted=is_sorted)

  def replace_reads(self, reads, is_sorted=False):
    """Replace the reads stored by this reader."""
    self.reads = reads
    self.is_sorted = is_sorted

  def iterate(self):
    """Iterate over all records in the reads.

    Returns:
      An iterator over nucleus.genomics.v1.Read's.
    """
    return self.reads

  def query(self, region):
    """Returns an iterator for going through the reads in the region.

    Args:
      region: nucleus.genomics.v1.Range. The query region.

    Returns:
      An iterator over nucleus.genomics.v1.Read protos.
    """
    # TODO(b/37353140): Add a faster query version for sorted reads.
    return (
        read for read in self.reads if utils.read_overlaps_region(read, region))



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for reading and writing FASTQ files.

The FASTQ format is described at
https://en.wikipedia.org/wiki/FASTQ_format

API for reading:

```python
from nucleus.io import fastq

with fastq.FastqReader(input_path) as reader:
  for record in reader:
    print(record)
```

where `record` is a `nucleus.genomics.v1.FastqRecord` protocol buffer.

API for writing:

```python
from nucleus.io import fastq

# records is an iterable of nucleus.genomics.v1.FastqRecord protocol buffers.
records = ...

with fastq.FastqWriter(output_path) as writer:
  for record in records:
    writer.write(record)
```

For both reading and writing, if the path provided to the constructor contains
'.tfrecord' as an extension, a `TFRecord` file is assumed and attempted to be
read or written. Otherwise, the filename is treated as a true FASTQ file.

Files that end in a '.gz' suffix cause the file to be treated as compressed
(with BGZF if it is a true FASTQ file, and with gzip if it is a TFRecord file).
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from nucleus.io import genomics_reader
from nucleus.io import genomics_writer
from nucleus.io.python import fastq_reader
from nucleus.io.python import fastq_writer
from nucleus.protos import fastq_pb2


class NativeFastqReader(genomics_reader.GenomicsReader):
  """Class for reading from native FASTQ files.

  Most users will want to use FastqReader instead, because it dynamically
  dispatches between reading native FASTQ files and TFRecord files based on the
  filename's extension.
  """

  def __init__(self, input_path):
    """Initializes a NativeFastqReader.

    Args:
      input_path: str. A path to a resource containing FASTQ records.
    """
    super(NativeFastqReader, self).__init__()

    fastq_path = input_path.encode('utf8')
    options = fastq_pb2.FastqReaderOptions()
    self._reader = fastq_reader.FastqReader.from_file(fastq_path, options)
    self.header = None

  def query(self, region):
    """Returns an iterator for going through the records in the region.

    NOTE: This function is not implemented by NativeFastqReader as there is no
    concept of genome ordering in the FASTQ format.
    """
    raise NotImplementedError('Can not query a FASTQ file')

  def iterate(self):
    """Returns an iterable of FastqRecord protos in the file."""
    return self._reader.iterate()

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._reader.__exit__(exit_type, exit_value, exit_traceback)


class FastqReader(genomics_reader.DispatchingGenomicsReader):
  """Class for reading FastqRecord protos from FASTQ or TFRecord files."""

  def _native_reader(self, input_path, **kwargs):
    return NativeFastqReader(input_path, **kwargs)

  def _record_proto(self):
    return fastq_pb2.FastqRecord


class NativeFastqWriter(genomics_writer.GenomicsWriter):
  """Class for writing to native FASTQ files.

  Most users will want FastqWriter, which will write to either native FASTQ
  files or TFRecord files, based on the output filename's extension.
  """

  def __init__(self, output_path, **kwargs):
    """Initializer for NativeFastqWriter.

    Args:
      output_path: str. The path to which to write the FASTQ file.
      **kwargs: optional arguments; presently ignored.
    """
    super(NativeFastqWriter, self).__init__()

    writer_options = fastq_pb2.FastqWriterOptions()
    self._writer = fastq_writer.FastqWriter.to_file(output_path, writer_options)

  def write(self, proto):
    self._writer.write(proto)

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._writer.__exit__(exit_type, exit_value, exit_traceback)


class FastqWriter(genomics_writer.DispatchingGenomicsWriter):
  """Class for writing FastqRecord protos to FASTQ or TFRecord files."""

  def _native_writer(self, output_path, **kwargs):
    return NativeFastqWriter(output_path, **kwargs)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Validates that a VCF file and a FASTA reference file correspond.

They correspond if:
a) they cover the same contigs,
b) the reference covers every variant in the vcf file, and
c) they agree on the reference bases covered by the variants.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from absl import app

from nucleus.io import fasta
from nucleus.io import vcf
from nucleus.util import variant_utils


def validate_contigs(ref_contigs, vcf_contigs):
  """Validate that the two lists of ContigInfos have the same set of names."""
  ref_names = {ci.name for ci in ref_contigs}
  vcf_names = {ci.name for ci in vcf_contigs}
  if ref_names != vcf_names:
    print('Contig names differ!')
    print('The following contigs are in one but not both: ')
    print(ref_names ^ vcf_names)
    sys.exit(-1)


def validate_variant(ref_reader, variant):
  """Validate that variant is covered by the reference and agrees with it."""
  var_range = variant_utils.variant_range(variant)
  ref_bases = ''
  try:
    ref_bases = ref_reader.query(var_range)
  except ValueError:
    print('Reference does not cover range {}: {}-{}'.format(
        var_range.reference_name, var_range.start, var_range.end))
    sys.exit(-1)

  if ref_bases != variant.reference_bases:
    print('In range {}:{}-{} '.format(
        var_range.reference_name, var_range.start, var_range.end))
    print('Reference says ', ref_bases)
    print('But variant says ', variant.reference_bases)
    sys.exit(-1)


def main(argv):
  if len(argv) != 3:
    print('Usage: {} <input_ref> <input_vcf>'.format(argv[0]))
    sys.exit(-1)
  in_ref = argv[1]
  in_vcf = argv[2]

  with fasta.IndexedFastaReader(in_ref) as ref_reader:
    with vcf.VcfReader(in_vcf) as vcf_reader:
      validate_contigs(ref_reader.header.contigs, vcf_reader.header.contigs)
      for variant in vcf_reader:
        validate_variant(ref_reader, variant)

  # VCF is valid!
  print('Reference and VCF are compatible.')
  sys.exit(0)


if __name__ == '__main__':
  app.run(main)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes that provide the interface for reading genomics data.

`GenomicsReader` defines the core API supported by readers, and is subclassed
directly or indirectly (via `DispatchingGenomicsReader`) for all concrete
implementations.

`TFRecordReader` is an implementation of the `GenomicsReader` API for reading
`TFRecord` files. This is usable for all data types when encoding data in
protocol buffers.

`DispatchingGenomicsReader` is an abstract class defined for convenience on top
of `GenomicsReader` that supports reading from either the native file format or
from `TFRecord` files of the corresponding protocol buffer used to encode data
of that file type. The input format assumed is dependent upon the filename of
the input data.

Concrete implementations for individual file types (e.g. BED, SAM, VCF, etc.)
reside in type-specific modules in this package. The instantiation of readers
may have reader-specific requirements documented there. General examples of the
`iterate()` and `query()` functionality are shown below.

```python
# Equivalent ways to iterate through all elements in a reader.
# 1. Using the reader itself as an iterable object.
kwargs = ...  # Reader-specific keyword arguments.
with GenomicsReaderSubClass(output_path, **kwargs) as reader:
  for proto in reader:
    do_something(reader.header, proto)

# 2. Calling the iterate() method of the reader explicitly.
with GenomicsReaderSubClass(output_path, **kwargs) as reader:
  for proto in reader.iterate():
    do_something(reader.header, proto)

# Querying for all elements within a specific region of the genome.
from nucleus.protos import range_pb2
region = range_pb2.Range(reference_name='chr1', start=10, end=20)

with GenomicsReaderSubClass(output_path, **kwargs) as reader:
  for proto in reader.query(region):
    do_something(reader.header, proto)
```
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import errno

from absl import logging
import six

from nucleus.io.python import tfrecord_reader


class GenomicsReader(six.Iterator):
  """Abstract base class for reading genomics data.

  In addition to the abstractmethods defined below, subclasses should
  also set a `header` member variable in their objects.
  """

  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def iterate(self):
    """Returns an iterator for going through all the file's records."""

  @abc.abstractmethod
  def query(self, region):
    """Returns an iterator for going through the records in the region.

    Args:
      region:  A nucleus.genomics.v1.Range.

    Returns:
      An iterator containing all and only records within the specified region.
    """

  def __enter__(self):
    """Enter a `with` block."""
    return self

  def __exit__(self, unused_type, unused_value, unused_traceback):
    """Exit a `with` block.  Typically, this will close the file."""

  def __init__(self):
    """Initializer."""
    # Some readers can only support one iterator at a time, so don't
    # create one now.  Rather, create it when needed in next().
    self.iterator = None

  def __iter__(self):
    """Allows users to use the object itself as an iterator."""
    return self.iterate()

  def __next__(self):
    """Allows users to use the object itself as an iterator."""
    if self.iterator is None:
      self.iterator = self.iterate()
    return six.next(self.iterator)


class TFRecordReader(GenomicsReader):
  """A GenomicsReader that reads protocol buffers from a TFRecord file.

  Example usage:
    reader = TFRecordReader('/tmp/my_file.tfrecords.gz',
                            proto=tensorflow.Example)
    for example in reader:
      process(example)

  Note that TFRecord files do not have headers, and do not need
  to be wrapped in a "with" block.
  """

  def __init__(self, input_path, proto, compression_type=None):
    """Initializes the TFRecordReader.

    Args:
      input_path:  The filename of the file to read.
      proto:  The protocol buffer type the TFRecord file is expected to
        contain.  For example, variants_pb2.Variant or reads_pb2.Read.
      compression_type:  Either 'ZLIB', 'GZIP', '' (uncompressed), or
        None.  If None, __init__ will guess the compression type based on
        the input_path's suffix.

    Raises:
      IOError: if there was any problem opening input_path for reading.
    """
    super(TFRecordReader, self).__init__()

    self.input_path = input_path
    self.proto = proto
    self.header = None

    if compression_type is None:
      compression_type = 'GZIP' if input_path.endswith('.gz') else ''

    self.reader = tfrecord_reader.TFRecordReader.from_file(
        input_path, compression_type)
    if self.reader is None:
      raise IOError(errno.EIO,
                    'Error trying to open %s for reading' % input_path)

  def iterate(self):
    """Returns an iterator for going through all the file's records."""
    while self.reader.get_next():
      yield self.proto.FromString(self.reader.get_record())

  def query(self, region):
    """Returns an iterator for going through the records in the region.

    NOTE: This function is not currently implemented by TFRecordReader as the
    TFRecord format does not provide a general mechanism for fast random access
    to elements in genome order.
    """
    raise NotImplementedError('Can not query TFRecord file')

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self.reader.close()

  @property
  def c_reader(self):
    """Returns the underlying C++ reader."""
    return self.reader


class DispatchingGenomicsReader(GenomicsReader):
  """A GenomicsReader that dispatches based on the file extension.

  If '.tfrecord' is present in the filename, a TFRecordReader is used.
  Otherwise, a native reader is.

  Subclasses of DispatchingGenomicsReader must define the following methods:
    * _native_reader()
    * _record_proto()
  """

  def __init__(self, input_path, **kwargs):
    super(DispatchingGenomicsReader, self).__init__()

    if '.tfrecord' in input_path:
      self._reader = TFRecordReader(
          input_path, proto=self._record_proto(),
          compression_type=kwargs.get('compression_type', None))
    else:
      # Remove compression_type, if present, from the arguments we pass to the
      # native reader.
      kwargs.pop('compression_type', None)
      self._reader = self._native_reader(input_path, **kwargs)
    logging.info('Reading %s with %s',
                 input_path, self._reader.__class__.__name__)
    self.header = getattr(self._reader, 'header', None)
    self._post_init_hook()

  @abc.abstractmethod
  def _native_reader(self, input_path, **kwargs):
    """Returns a GenomicsReader for reading the records `natively`.

    Args:
      input_path: The path to the native file to read.
      **kwargs:  Zero or more keyword arguments.

    Returns:
      A GenomicsReader.
    """

  @abc.abstractmethod
  def _record_proto(self):
    """Returns the protocol buffer type used by this reader."""

  def iterate(self):
    return self._reader.iterate()

  def query(self, region):
    return self._reader.query(region)

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._reader.__exit__(exit_type, exit_value, exit_traceback)

  def _post_init_hook(self):
    """Hook for subclasses to run code at the end of __init__."""



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CLIF postprocessors."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import six

from nucleus.protos import bed_pb2
from nucleus.protos import bedgraph_pb2
from nucleus.protos import fastq_pb2
from nucleus.protos import gff_pb2
from nucleus.protos import reads_pb2
from nucleus.protos import variants_pb2


def ValueErrorOnFalse(ok, *args):
  """Returns None / arg / (args,...) if ok."""
  if not isinstance(ok, bool):
    raise TypeError('Use ValueErrorOnFalse only on bool return value')
  if not ok:
    raise ValueError('CLIF wrapped call returned False')
  # Plain return args will turn 1 into (1,)  and None into () which is unwanted.
  if args:
    return args if len(args) > 1 else args[0]
  return None


class WrappedCppIterable(six.Iterator):
  """This class gives Python iteration semantics on top of a C++ 'Iterable'."""

  __metaclass__ = abc.ABCMeta

  def __init__(self, cc_iterable):
    self._cc_iterable = cc_iterable

  def __enter__(self):
    self._cc_iterable.__enter__()
    return self

  def __exit__(self, type_, value, traceback):
    self._cc_iterable.__exit__(type_, value, traceback)

  def __iter__(self):
    return self

  @abc.abstractmethod
  def _raw_next(self):
    """Sub-classes should implement __next__ in this method."""

  def __next__(self):
    try:
      record, not_done = self._raw_next()
    except AttributeError:
      if self._cc_iterable is None:
        raise ValueError('No underlying iterable. This may occur if trying to '
                         'create multiple concurrent iterators from the same '
                         'reader. Try wrapping your call to the iterator in a '
                         '`with` block or materializing the entire iterable '
                         'explicitly.')
      else:
        raise
    if not_done:
      return record
    else:
      raise StopIteration


class WrappedBedIterable(WrappedCppIterable):

  def _raw_next(self):
    record = bed_pb2.BedRecord()
    not_done = self._cc_iterable.PythonNext(record)
    return record, not_done


class WrappedBedGraphIterable(WrappedCppIterable):

  def _raw_next(self):
    record = bedgraph_pb2.BedGraphRecord()
    not_done = self._cc_iterable.PythonNext(record)
    return record, not_done


class WrappedFastqIterable(WrappedCppIterable):

  def _raw_next(self):
    record = fastq_pb2.FastqRecord()
    not_done = self._cc_iterable.PythonNext(record)
    return record, not_done


class WrappedGffIterable(WrappedCppIterable):

  def _raw_next(self):
    record = gff_pb2.GffRecord()
    not_done = self._cc_iterable.PythonNext(record)
    return record, not_done


class WrappedReferenceIterable(WrappedCppIterable):

  def _raw_next(self):
    not_done, record = self._cc_iterable.Next()
    return record, not_done


class WrappedSamIterable(WrappedCppIterable):

  def _raw_next(self):
    record = reads_pb2.Read()
    not_done = self._cc_iterable.PythonNext(record)
    return record, not_done


class WrappedVariantIterable(WrappedCppIterable):

  def _raw_next(self):
    record = variants_pb2.Variant()
    not_done = self._cc_iterable.PythonNext(record)
    return record, not_done



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Writes all the variants in a VCF file with a quality greater than 3.01."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from absl import app

from nucleus.io import vcf


def main(argv):
  if len(argv) != 3:
    print('Usage: {} <input_vcf> <output_vcf>'.format(argv[0]))
    sys.exit(-1)
  in_vcf = argv[1]
  out_vcf = argv[2]

  # Please try to keep the following part in sync with the documenation in
  # g3doc/overview.md.
  with vcf.VcfReader(in_vcf) as reader:
    print('Sample names in VCF: ', ' '.join(reader.header.sample_names))
    with vcf.VcfWriter(out_vcf, header=reader.header) as writer:
      for variant in reader:
        if variant.quality > 3.01:
          writer.write(variant)


if __name__ == '__main__':
  app.run(main)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nucleus/protos/reads.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nucleus.protos import cigar_pb2 as nucleus_dot_protos_dot_cigar__pb2
from nucleus.protos import position_pb2 as nucleus_dot_protos_dot_position__pb2
from nucleus.protos import reference_pb2 as nucleus_dot_protos_dot_reference__pb2
from nucleus.protos import struct_pb2 as nucleus_dot_protos_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='nucleus/protos/reads.proto',
  package='nucleus.genomics.v1',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1anucleus/protos/reads.proto\x12\x13nucleus.genomics.v1\x1a\x1anucleus/protos/cigar.proto\x1a\x1dnucleus/protos/position.proto\x1a\x1enucleus/protos/reference.proto\x1a\x1bnucleus/protos/struct.proto\"\x8a\x01\n\x0fLinearAlignment\x12/\n\x08position\x18\x01 \x01(\x0b\x32\x1d.nucleus.genomics.v1.Position\x12\x17\n\x0fmapping_quality\x18\x02 \x01(\x05\x12-\n\x05\x63igar\x18\x03 \x03(\x0b\x32\x1e.nucleus.genomics.v1.CigarUnit\"\xe0\x04\n\x04Read\x12\n\n\x02id\x18\x01 \x01(\t\x12\x15\n\rread_group_id\x18\x02 \x01(\t\x12\x19\n\x11read_group_set_id\x18\x03 \x01(\t\x12\x15\n\rfragment_name\x18\x04 \x01(\t\x12\x18\n\x10proper_placement\x18\x05 \x01(\x08\x12\x1a\n\x12\x64uplicate_fragment\x18\x06 \x01(\x08\x12\x17\n\x0f\x66ragment_length\x18\x07 \x01(\x05\x12\x13\n\x0bread_number\x18\x08 \x01(\x05\x12\x14\n\x0cnumber_reads\x18\t \x01(\x05\x12$\n\x1c\x66\x61iled_vendor_quality_checks\x18\n \x01(\x08\x12\x37\n\talignment\x18\x0b \x01(\x0b\x32$.nucleus.genomics.v1.LinearAlignment\x12\x1b\n\x13secondary_alignment\x18\x0c \x01(\x08\x12\x1f\n\x17supplementary_alignment\x18\r \x01(\x08\x12\x18\n\x10\x61ligned_sequence\x18\x0e \x01(\t\x12\x17\n\x0f\x61ligned_quality\x18\x0f \x03(\x05\x12\x39\n\x12next_mate_position\x18\x10 \x01(\x0b\x32\x1d.nucleus.genomics.v1.Position\x12\x31\n\x04info\x18\x11 \x03(\x0b\x32#.nucleus.genomics.v1.Read.InfoEntry\x1aK\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12-\n\x05value\x18\x02 \x01(\x0b\x32\x1e.nucleus.genomics.v1.ListValue:\x02\x38\x01\"\xe1\x03\n\tSamHeader\x12\x16\n\x0e\x66ormat_version\x18\x01 \x01(\t\x12\x42\n\rsorting_order\x18\x02 \x01(\x0e\x32+.nucleus.genomics.v1.SamHeader.SortingOrder\x12L\n\x12\x61lignment_grouping\x18\x03 \x01(\x0e\x32\x30.nucleus.genomics.v1.SamHeader.AlignmentGrouping\x12\x30\n\x07\x63ontigs\x18\x04 \x03(\x0b\x32\x1f.nucleus.genomics.v1.ContigInfo\x12\x33\n\x0bread_groups\x18\x05 \x03(\x0b\x32\x1e.nucleus.genomics.v1.ReadGroup\x12.\n\x08programs\x18\x06 \x03(\x0b\x32\x1c.nucleus.genomics.v1.Program\x12\x10\n\x08\x63omments\x18\x07 \x03(\t\"H\n\x0cSortingOrder\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0c\n\x08UNSORTED\x10\x01\x12\r\n\tQUERYNAME\x10\x02\x12\x0e\n\nCOORDINATE\x10\x03\"7\n\x11\x41lignmentGrouping\x12\x08\n\x04NONE\x10\x00\x12\t\n\x05QUERY\x10\x01\x12\r\n\tREFERENCE\x10\x02\"\x9d\x02\n\tReadGroup\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x19\n\x11sequencing_center\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0c\n\x04\x64\x61te\x18\x04 \x01(\t\x12\x12\n\nflow_order\x18\x05 \x01(\t\x12\x14\n\x0ckey_sequence\x18\x06 \x01(\t\x12\x12\n\nlibrary_id\x18\x07 \x01(\t\x12\x13\n\x0bprogram_ids\x18\x08 \x03(\t\x12\x1d\n\x15predicted_insert_size\x18\t \x01(\x05\x12\x10\n\x08platform\x18\n \x01(\t\x12\x16\n\x0eplatform_model\x18\x0b \x01(\t\x12\x15\n\rplatform_unit\x18\x0c \x01(\t\x12\x11\n\tsample_id\x18\r \x01(\t\"x\n\x07Program\x12\n\n\x02id\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x14\n\x0c\x63ommand_line\x18\x01 \x01(\t\x12\x17\n\x0fprev_program_id\x18\x04 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x06 \x01(\t\x12\x0f\n\x07version\x18\x05 \x01(\t\"\xf0\x02\n\x10SamReaderOptions\x12@\n\x11read_requirements\x18\x01 \x01(\x0b\x32%.nucleus.genomics.v1.ReadRequirements\x12R\n\x12\x61ux_field_handling\x18\x03 \x01(\x0e\x32\x36.nucleus.genomics.v1.SamReaderOptions.AuxFieldHandling\x12\x16\n\x0ehts_block_size\x18\x04 \x01(\x03\x12\x1b\n\x13\x64ownsample_fraction\x18\x05 \x01(\x02\x12\x13\n\x0brandom_seed\x18\x06 \x01(\x03\x12(\n use_original_base_quality_scores\x18\n \x01(\x08\"R\n\x10\x41uxFieldHandling\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x13\n\x0fSKIP_AUX_FIELDS\x10\x01\x12\x18\n\x14PARSE_ALL_AUX_FIELDS\x10\x02\"\xa7\x03\n\x10ReadRequirements\x12\x17\n\x0fkeep_duplicates\x18\x01 \x01(\x08\x12)\n!keep_failed_vendor_quality_checks\x18\x02 \x01(\x08\x12!\n\x19keep_secondary_alignments\x18\x03 \x01(\x08\x12%\n\x1dkeep_supplementary_alignments\x18\x04 \x01(\x08\x12\x16\n\x0ekeep_unaligned\x18\x05 \x01(\x08\x12\x1e\n\x16keep_improperly_placed\x18\x06 \x01(\x08\x12\x1b\n\x13min_mapping_quality\x18\x07 \x01(\x05\x12\x18\n\x10min_base_quality\x18\x08 \x01(\x05\x12W\n\x15min_base_quality_mode\x18\t \x01(\x0e\x32\x38.nucleus.genomics.v1.ReadRequirements.MinBaseQualityMode\"=\n\x12MinBaseQualityMode\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x16\n\x12\x45NFORCED_BY_CLIENT\x10\x01\x62\x06proto3')
  ,
  dependencies=[nucleus_dot_protos_dot_cigar__pb2.DESCRIPTOR,nucleus_dot_protos_dot_position__pb2.DESCRIPTOR,nucleus_dot_protos_dot_reference__pb2.DESCRIPTOR,nucleus_dot_protos_dot_struct__pb2.DESCRIPTOR,])



_SAMHEADER_SORTINGORDER = _descriptor.EnumDescriptor(
  name='SortingOrder',
  full_name='nucleus.genomics.v1.SamHeader.SortingOrder',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNSORTED', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='QUERYNAME', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COORDINATE', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1276,
  serialized_end=1348,
)
_sym_db.RegisterEnumDescriptor(_SAMHEADER_SORTINGORDER)

_SAMHEADER_ALIGNMENTGROUPING = _descriptor.EnumDescriptor(
  name='AlignmentGrouping',
  full_name='nucleus.genomics.v1.SamHeader.AlignmentGrouping',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NONE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='QUERY', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REFERENCE', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1350,
  serialized_end=1405,
)
_sym_db.RegisterEnumDescriptor(_SAMHEADER_ALIGNMENTGROUPING)

_SAMREADEROPTIONS_AUXFIELDHANDLING = _descriptor.EnumDescriptor(
  name='AuxFieldHandling',
  full_name='nucleus.genomics.v1.SamReaderOptions.AuxFieldHandling',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SKIP_AUX_FIELDS', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PARSE_ALL_AUX_FIELDS', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2104,
  serialized_end=2186,
)
_sym_db.RegisterEnumDescriptor(_SAMREADEROPTIONS_AUXFIELDHANDLING)

_READREQUIREMENTS_MINBASEQUALITYMODE = _descriptor.EnumDescriptor(
  name='MinBaseQualityMode',
  full_name='nucleus.genomics.v1.ReadRequirements.MinBaseQualityMode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENFORCED_BY_CLIENT', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2551,
  serialized_end=2612,
)
_sym_db.RegisterEnumDescriptor(_READREQUIREMENTS_MINBASEQUALITYMODE)


_LINEARALIGNMENT = _descriptor.Descriptor(
  name='LinearAlignment',
  full_name='nucleus.genomics.v1.LinearAlignment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='position', full_name='nucleus.genomics.v1.LinearAlignment.position', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mapping_quality', full_name='nucleus.genomics.v1.LinearAlignment.mapping_quality', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cigar', full_name='nucleus.genomics.v1.LinearAlignment.cigar', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=172,
  serialized_end=310,
)


_READ_INFOENTRY = _descriptor.Descriptor(
  name='InfoEntry',
  full_name='nucleus.genomics.v1.Read.InfoEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='nucleus.genomics.v1.Read.InfoEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='nucleus.genomics.v1.Read.InfoEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=846,
  serialized_end=921,
)

_READ = _descriptor.Descriptor(
  name='Read',
  full_name='nucleus.genomics.v1.Read',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='nucleus.genomics.v1.Read.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='read_group_id', full_name='nucleus.genomics.v1.Read.read_group_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='read_group_set_id', full_name='nucleus.genomics.v1.Read.read_group_set_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fragment_name', full_name='nucleus.genomics.v1.Read.fragment_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proper_placement', full_name='nucleus.genomics.v1.Read.proper_placement', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='duplicate_fragment', full_name='nucleus.genomics.v1.Read.duplicate_fragment', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fragment_length', full_name='nucleus.genomics.v1.Read.fragment_length', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='read_number', full_name='nucleus.genomics.v1.Read.read_number', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='number_reads', full_name='nucleus.genomics.v1.Read.number_reads', index=8,
      number=9, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='failed_vendor_quality_checks', full_name='nucleus.genomics.v1.Read.failed_vendor_quality_checks', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='alignment', full_name='nucleus.genomics.v1.Read.alignment', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='secondary_alignment', full_name='nucleus.genomics.v1.Read.secondary_alignment', index=11,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supplementary_alignment', full_name='nucleus.genomics.v1.Read.supplementary_alignment', index=12,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aligned_sequence', full_name='nucleus.genomics.v1.Read.aligned_sequence', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aligned_quality', full_name='nucleus.genomics.v1.Read.aligned_quality', index=14,
      number=15, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='next_mate_position', full_name='nucleus.genomics.v1.Read.next_mate_position', index=15,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='info', full_name='nucleus.genomics.v1.Read.info', index=16,
      number=17, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_READ_INFOENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=313,
  serialized_end=921,
)


_SAMHEADER = _descriptor.Descriptor(
  name='SamHeader',
  full_name='nucleus.genomics.v1.SamHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='format_version', full_name='nucleus.genomics.v1.SamHeader.format_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sorting_order', full_name='nucleus.genomics.v1.SamHeader.sorting_order', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='alignment_grouping', full_name='nucleus.genomics.v1.SamHeader.alignment_grouping', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='contigs', full_name='nucleus.genomics.v1.SamHeader.contigs', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='read_groups', full_name='nucleus.genomics.v1.SamHeader.read_groups', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='programs', full_name='nucleus.genomics.v1.SamHeader.programs', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='comments', full_name='nucleus.genomics.v1.SamHeader.comments', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SAMHEADER_SORTINGORDER,
    _SAMHEADER_ALIGNMENTGROUPING,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=924,
  serialized_end=1405,
)


_READGROUP = _descriptor.Descriptor(
  name='ReadGroup',
  full_name='nucleus.genomics.v1.ReadGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='nucleus.genomics.v1.ReadGroup.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sequencing_center', full_name='nucleus.genomics.v1.ReadGroup.sequencing_center', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='nucleus.genomics.v1.ReadGroup.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='date', full_name='nucleus.genomics.v1.ReadGroup.date', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='flow_order', full_name='nucleus.genomics.v1.ReadGroup.flow_order', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='key_sequence', full_name='nucleus.genomics.v1.ReadGroup.key_sequence', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='library_id', full_name='nucleus.genomics.v1.ReadGroup.library_id', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='program_ids', full_name='nucleus.genomics.v1.ReadGroup.program_ids', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='predicted_insert_size', full_name='nucleus.genomics.v1.ReadGroup.predicted_insert_size', index=8,
      number=9, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='platform', full_name='nucleus.genomics.v1.ReadGroup.platform', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='platform_model', full_name='nucleus.genomics.v1.ReadGroup.platform_model', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='platform_unit', full_name='nucleus.genomics.v1.ReadGroup.platform_unit', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sample_id', full_name='nucleus.genomics.v1.ReadGroup.sample_id', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1408,
  serialized_end=1693,
)


_PROGRAM = _descriptor.Descriptor(
  name='Program',
  full_name='nucleus.genomics.v1.Program',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='nucleus.genomics.v1.Program.id', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='nucleus.genomics.v1.Program.name', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='command_line', full_name='nucleus.genomics.v1.Program.command_line', index=2,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='prev_program_id', full_name='nucleus.genomics.v1.Program.prev_program_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='nucleus.genomics.v1.Program.description', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='nucleus.genomics.v1.Program.version', index=5,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1695,
  serialized_end=1815,
)


_SAMREADEROPTIONS = _descriptor.Descriptor(
  name='SamReaderOptions',
  full_name='nucleus.genomics.v1.SamReaderOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='read_requirements', full_name='nucleus.genomics.v1.SamReaderOptions.read_requirements', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aux_field_handling', full_name='nucleus.genomics.v1.SamReaderOptions.aux_field_handling', index=1,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hts_block_size', full_name='nucleus.genomics.v1.SamReaderOptions.hts_block_size', index=2,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='downsample_fraction', full_name='nucleus.genomics.v1.SamReaderOptions.downsample_fraction', index=3,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='random_seed', full_name='nucleus.genomics.v1.SamReaderOptions.random_seed', index=4,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='use_original_base_quality_scores', full_name='nucleus.genomics.v1.SamReaderOptions.use_original_base_quality_scores', index=5,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SAMREADEROPTIONS_AUXFIELDHANDLING,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1818,
  serialized_end=2186,
)


_READREQUIREMENTS = _descriptor.Descriptor(
  name='ReadRequirements',
  full_name='nucleus.genomics.v1.ReadRequirements',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='keep_duplicates', full_name='nucleus.genomics.v1.ReadRequirements.keep_duplicates', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keep_failed_vendor_quality_checks', full_name='nucleus.genomics.v1.ReadRequirements.keep_failed_vendor_quality_checks', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keep_secondary_alignments', full_name='nucleus.genomics.v1.ReadRequirements.keep_secondary_alignments', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keep_supplementary_alignments', full_name='nucleus.genomics.v1.ReadRequirements.keep_supplementary_alignments', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keep_unaligned', full_name='nucleus.genomics.v1.ReadRequirements.keep_unaligned', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keep_improperly_placed', full_name='nucleus.genomics.v1.ReadRequirements.keep_improperly_placed', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='min_mapping_quality', full_name='nucleus.genomics.v1.ReadRequirements.min_mapping_quality', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='min_base_quality', full_name='nucleus.genomics.v1.ReadRequirements.min_base_quality', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='min_base_quality_mode', full_name='nucleus.genomics.v1.ReadRequirements.min_base_quality_mode', index=8,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _READREQUIREMENTS_MINBASEQUALITYMODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2189,
  serialized_end=2612,
)

_LINEARALIGNMENT.fields_by_name['position'].message_type = nucleus_dot_protos_dot_position__pb2._POSITION
_LINEARALIGNMENT.fields_by_name['cigar'].message_type = nucleus_dot_protos_dot_cigar__pb2._CIGARUNIT
_READ_INFOENTRY.fields_by_name['value'].message_type = nucleus_dot_protos_dot_struct__pb2._LISTVALUE
_READ_INFOENTRY.containing_type = _READ
_READ.fields_by_name['alignment'].message_type = _LINEARALIGNMENT
_READ.fields_by_name['next_mate_position'].message_type = nucleus_dot_protos_dot_position__pb2._POSITION
_READ.fields_by_name['info'].message_type = _READ_INFOENTRY
_SAMHEADER.fields_by_name['sorting_order'].enum_type = _SAMHEADER_SORTINGORDER
_SAMHEADER.fields_by_name['alignment_grouping'].enum_type = _SAMHEADER_ALIGNMENTGROUPING
_SAMHEADER.fields_by_name['contigs'].message_type = nucleus_dot_protos_dot_reference__pb2._CONTIGINFO
_SAMHEADER.fields_by_name['read_groups'].message_type = _READGROUP
_SAMHEADER.fields_by_name['programs'].message_type = _PROGRAM
_SAMHEADER_SORTINGORDER.containing_type = _SAMHEADER
_SAMHEADER_ALIGNMENTGROUPING.containing_type = _SAMHEADER
_SAMREADEROPTIONS.fields_by_name['read_requirements'].message_type = _READREQUIREMENTS
_SAMREADEROPTIONS.fields_by_name['aux_field_handling'].enum_type = _SAMREADEROPTIONS_AUXFIELDHANDLING
_SAMREADEROPTIONS_AUXFIELDHANDLING.containing_type = _SAMREADEROPTIONS
_READREQUIREMENTS.fields_by_name['min_base_quality_mode'].enum_type = _READREQUIREMENTS_MINBASEQUALITYMODE
_READREQUIREMENTS_MINBASEQUALITYMODE.containing_type = _READREQUIREMENTS
DESCRIPTOR.message_types_by_name['LinearAlignment'] = _LINEARALIGNMENT
DESCRIPTOR.message_types_by_name['Read'] = _READ
DESCRIPTOR.message_types_by_name['SamHeader'] = _SAMHEADER
DESCRIPTOR.message_types_by_name['ReadGroup'] = _READGROUP
DESCRIPTOR.message_types_by_name['Program'] = _PROGRAM
DESCRIPTOR.message_types_by_name['SamReaderOptions'] = _SAMREADEROPTIONS
DESCRIPTOR.message_types_by_name['ReadRequirements'] = _READREQUIREMENTS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LinearAlignment = _reflection.GeneratedProtocolMessageType('LinearAlignment', (_message.Message,), {
  'DESCRIPTOR' : _LINEARALIGNMENT,
  '__module__' : 'nucleus.protos.reads_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.LinearAlignment)
  })
_sym_db.RegisterMessage(LinearAlignment)

Read = _reflection.GeneratedProtocolMessageType('Read', (_message.Message,), {

  'InfoEntry' : _reflection.GeneratedProtocolMessageType('InfoEntry', (_message.Message,), {
    'DESCRIPTOR' : _READ_INFOENTRY,
    '__module__' : 'nucleus.protos.reads_pb2'
    # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.Read.InfoEntry)
    })
  ,
  'DESCRIPTOR' : _READ,
  '__module__' : 'nucleus.protos.reads_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.Read)
  })
_sym_db.RegisterMessage(Read)
_sym_db.RegisterMessage(Read.InfoEntry)

SamHeader = _reflection.GeneratedProtocolMessageType('SamHeader', (_message.Message,), {
  'DESCRIPTOR' : _SAMHEADER,
  '__module__' : 'nucleus.protos.reads_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.SamHeader)
  })
_sym_db.RegisterMessage(SamHeader)

ReadGroup = _reflection.GeneratedProtocolMessageType('ReadGroup', (_message.Message,), {
  'DESCRIPTOR' : _READGROUP,
  '__module__' : 'nucleus.protos.reads_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.ReadGroup)
  })
_sym_db.RegisterMessage(ReadGroup)

Program = _reflection.GeneratedProtocolMessageType('Program', (_message.Message,), {
  'DESCRIPTOR' : _PROGRAM,
  '__module__' : 'nucleus.protos.reads_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.Program)
  })
_sym_db.RegisterMessage(Program)

SamReaderOptions = _reflection.GeneratedProtocolMessageType('SamReaderOptions', (_message.Message,), {
  'DESCRIPTOR' : _SAMREADEROPTIONS,
  '__module__' : 'nucleus.protos.reads_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.SamReaderOptions)
  })
_sym_db.RegisterMessage(SamReaderOptions)

ReadRequirements = _reflection.GeneratedProtocolMessageType('ReadRequirements', (_message.Message,), {
  'DESCRIPTOR' : _READREQUIREMENTS,
  '__module__' : 'nucleus.protos.reads_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.ReadRequirements)
  })
_sym_db.RegisterMessage(ReadRequirements)


_READ_INFOENTRY._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility library for working with protobufs."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from google.protobuf.internal import api_implementation


def uses_fast_cpp_protos_or_die():
  """Raises an error if a slow protobuf implementation is being used."""
  if api_implementation.Type() != 'cpp':
    raise ValueError('Expected to be using C++ protobuf implementation '
                     '(api_implementation.Type() == "cpp") but it is {}'.format(
                         api_implementation.Type()))



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for reading and writing GFF files.

The GFF format is described at
https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md.

API for reading:

```python
from nucleus.io import gff

# Iterate through all records.
with gff.GffReader(input_path) as reader:
  for record in reader:
    print(record)
```

where `record` is a `nucleus.genomics.v1.GffRecord` protocol buffer.

API for writing:

```python
from nucleus.io import gff
from nucleus.protos import gff_pb2

# records is an iterable of nucleus.genomics.v1.GffRecord protocol buffers.
records = ...

header = gff_pb2.GffHeader()

# Write all records to the desired output path.
with gff.GffWriter(output_path, header) as writer:
  for record in records:
    writer.write(record)
```

For both reading and writing, if the path provided to the constructor contains
'.tfrecord' as an extension, a `TFRecord` file is assumed and attempted to be
read or written. Otherwise, the filename is treated as a true GFF file.

Files that end in a '.gz' suffix cause the file to be treated as compressed
(with BGZF if it is a true GFF file, and with gzip if it is a TFRecord file).
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from nucleus.io import genomics_reader
from nucleus.io import genomics_writer
from nucleus.io.python import gff_reader
from nucleus.io.python import gff_writer
from nucleus.protos import gff_pb2


class NativeGffReader(genomics_reader.GenomicsReader):
  """Class for reading from native GFF files.

  Most users will want to use GffReader instead, because it dynamically
  dispatches between reading native GFF files and TFRecord files based on the
  filename's extension.
  """

  def __init__(self, input_path):
    """Initializes a NativeGffReader.

    Args:
      input_path: string. A path to a resource containing GFF records.
    """
    super(NativeGffReader, self).__init__()
    gff_path = input_path.encode('utf8')
    reader_options = gff_pb2.GffReaderOptions()
    self._reader = gff_reader.GffReader.from_file(gff_path, reader_options)
    self.header = self._reader.header

  def query(self):
    """Returns an iterator for going through the records in the region.

    NOTE: This function is not currently implemented by NativeGffReader though
    it could be implemented for sorted, tabix-indexed GFF files.
    """
    raise NotImplementedError('Can not currently query a GFF file')

  def iterate(self):
    """Returns an iterable of GffRecord protos in the file."""
    return self._reader.iterate()

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._reader.__exit__(exit_type, exit_value, exit_traceback)


class GffReader(genomics_reader.DispatchingGenomicsReader):
  """Class for reading GffRecord protos from GFF or TFRecord files."""

  def _native_reader(self, input_path, **kwargs):
    return NativeGffReader(input_path, **kwargs)

  def _record_proto(self):
    return gff_pb2.GffRecord


class NativeGffWriter(genomics_writer.GenomicsWriter):
  """Class for writing to native GFF files.

  Most users will want GffWriter, which will write to either native GFF
  files or TFRecord files, based on the output filename's extension.
  """

  def __init__(self, output_path, header):
    """Initializer for NativeGffWriter.

    Args:
      output_path: str. The path to which to write the GFF file.
      header: nucleus.genomics.v1.GffHeader. The header that defines all
        information germane to the constituent GFF records.
    """
    super(NativeGffWriter, self).__init__()
    writer_options = gff_pb2.GffWriterOptions()
    self._writer = gff_writer.GffWriter.to_file(output_path, header,
                                                writer_options)

  def write(self, proto):
    self._writer.write(proto)

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._writer.__exit__(exit_type, exit_value, exit_traceback)


class GffWriter(genomics_writer.DispatchingGenomicsWriter):
  """Class for writing GffRecord protos to GFF or TFRecord files."""

  def _native_writer(self, output_path, header):
    return NativeGffWriter(output_path, header=header)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""I/O for TFRecord files.

Utilities for reading and writing TFRecord files, especially those containing
serialized TensorFlow Example protocol buffers.
"""

# Important: Please keep this module free of TensorFlow C++ extensions.
# This makes it easy to build pure python packages for training that work with
# CMLE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import heapq

import contextlib2

from nucleus.io import genomics_reader
from nucleus.io import genomics_writer
from nucleus.io import sharded_file_utils
from nucleus.protos import example_pb2


# pylint: disable=invalid-name
def Reader(path, proto=None, compression_type=None):
  """A TFRecordReader that defaults to tf.Example protos."""
  if not proto:
    proto = example_pb2.Example

  return genomics_reader.TFRecordReader(
      path, proto, compression_type=compression_type)


def Writer(path, compression_type=None):
  """A convenience wrapper around genomics_writer.TFRecordWriter."""
  return genomics_writer.TFRecordWriter(
      path, compression_type=compression_type)
# pylint: enable=invalid-name


# TODO(thomaswc): Refactor all of the following (b/128406743).
def read_tfrecords(path, proto=None, max_records=None, compression_type=None):
  """Yields the parsed records in a TFRecord file path.

  Note that path can be sharded filespec (path@N) in which case this function
  will read each shard in order; i.e. shard 0 will read each entry in order,
  then shard 1, ...

  Args:
    path: String. A path to a TFRecord file containing protos.
    proto: A proto class. proto.FromString() will be called on each serialized
      record in path to parse it.
    max_records: int >= 0 or None. Maximum number of records to read from path.
      If None, the default, all records will be read.
    compression_type: 'GZIP', 'ZLIB', '' (uncompressed), or None to autodetect
      based on file extension.

  Yields:
    proto.FromString() values on each record in path in order.
  """
  if sharded_file_utils.is_sharded_file_spec(path):
    paths = sharded_file_utils.generate_sharded_filenames(path)
  else:
    paths = [path]

  i = 0
  for path in paths:
    with Reader(path, proto, compression_type=compression_type) as reader:
      for record in reader.iterate():
        i += 1
        if max_records is not None and i > max_records:
          return
        yield record


def read_shard_sorted_tfrecords(path,
                                key,
                                proto=None,
                                max_records=None,
                                compression_type=None):
  """Yields the parsed records in a TFRecord file path in sorted order.

  The input TFRecord file must have each shard already in sorted order when
  using the key function for comparison (but elements can be interleaved across
  shards). Under those constraints, the elements will be yielded in a global
  sorted order.

  Args:
    path: String. A path to a TFRecord-formatted file containing protos.
    key: Callable. A function that takes as input a single instance of the proto
      class and returns a value on which the comparison for sorted ordering is
      performed.
    proto: A proto class. proto.FromString() will be called on each serialized
      record in path to parse it.
    max_records: int >= 0 or None. Maximum number of records to read from path.
      If None, the default, all records will be read.
    compression_type: 'GZIP', 'ZLIB', '' (uncompressed), or None to autodetect
      based on file extension.

  Yields:
    proto.FromString() values on each record in path in sorted order.
  """
  if sharded_file_utils.is_sharded_file_spec(path):
    paths = sharded_file_utils.generate_sharded_filenames(path)
  else:
    paths = [path]

  keyed_iterables = []
  for path in paths:
    protos = Reader(path, proto, compression_type=compression_type).iterate()
    keyed_iterables.append(((key(elem), elem) for elem in protos))

  for i, (_, value) in enumerate(heapq.merge(*keyed_iterables)):
    if max_records is not None and i >= max_records:
      return
    yield value


def write_tfrecords(protos, output_path, compression_type=None):
  """Writes protos to output_path.

  This function writes serialized strings of each proto in protos to output_path
  in their original order. If output_path is a sharded file (e.g., foo@2), this
  function will write the protos spread out as evenly as possible among the
  individual components of the sharded spec (e.g., foo-00000-of-00002 and
  foo-00001-of-00002). Note that the order of records in the sharded files may
  differ from the order in protos due to the striping.

  Args:
    protos: An iterable of protobufs. The objects we want to write out.
    output_path: str. The filepath where we want to write protos.
    compression_type: 'GZIP', 'ZLIB', '' (uncompressed), or None to autodetect
      based on file extension.
  """
  if sharded_file_utils.is_sharded_file_spec(output_path):
    with contextlib2.ExitStack() as stack:
      _, n_shards, _ = sharded_file_utils.parse_sharded_file_spec(output_path)
      writers = [
          stack.enter_context(
              Writer(sharded_file_utils.sharded_filename(
                  output_path, i), compression_type=compression_type))
          for i in range(n_shards)
      ]
      for i, proto in enumerate(protos):
        writers[i % n_shards].write(proto)
  else:
    with Writer(output_path, compression_type=compression_type) as writer:
      for proto in protos:
        writer.write(proto)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for reading and writing BedGraph files.

The BedGraph format is described at
https://genome.ucsc.edu/goldenpath/help/bedgraph.html

API for reading:

```python
from nucleus.io import bedgraph

# Iterate through all records.
with bed.BedGraphReader(input_path) as reader:
  for record in reader:
    print(record)
```

where `record` is a `nucleus.genomics.v1.BedGraphRecord` protocol buffer.

API for writing:

```python
from nucleus.io import bedgraph
from nucleus.protos import bedgraph_pb2

# records is an iterable of nucleus.genomics.v1.BedGraphRecord protocol buffers.
records = ...

# Write all records to the desired output path.
with bed.BedGraphWriter(output_path) as writer:
  for record in records:
    writer.write(record)
```

For both reading and writing, if the path provided to the constructor contains
'.tfrecord' as an extension, a `TFRecord` file is assumed and attempted to be
read or written. Otherwise, the filename is treated as a true BedGraph file.

Files that end in a '.gz' suffix cause the file to be treated as compressed
(with BGZF if it is a BedGraph file, and with gzip if it is a TFRecord file).
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from nucleus.io import genomics_reader
from nucleus.io import genomics_writer
from nucleus.io.python import bedgraph_reader
from nucleus.io.python import bedgraph_writer
from nucleus.protos import bedgraph_pb2


class NativeBedGraphReader(genomics_reader.GenomicsReader):
  """Class for reading from native BedGraph files.

  Most users will want to use BedGraphReader instead, because it dynamically
  dispatches between reading native BedGraph files and TFRecord files based on
  the filename's extension.
  """

  def __init__(self, input_path, num_fields=0):
    """Initializes a NativeBedGraphReader.

    Args:
      input_path: string. A path to a resource containing BedGraph records.
      num_fields: int. The number of fields to read in the BedGraph. If unset or
        set to zero, all fields in the input are read.
    """
    super(NativeBedGraphReader, self).__init__()

    bedgraph_path = input_path.encode('utf8')
    self._reader = bedgraph_reader.BedGraphReader.from_file(bedgraph_path)

  def query(self):
    """Returns an iterator for going through the records in the region.

    NOTE: This function is not currently implemented by NativeBedGraphReader
    though it could be implemented for sorted, tabix-indexed BedGraph files.
    """
    raise NotImplementedError('Can not currently query a BedGraph file')

  def iterate(self):
    """Returns an iterable of BedGraphRecord protos in the file."""
    return self._reader.iterate()

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._reader.__exit__(exit_type, exit_value, exit_traceback)


class BedGraphReader(genomics_reader.DispatchingGenomicsReader):
  """Class for reading BedGraphRecord protos from BedGraph or TFRecord files."""

  def _native_reader(self, input_path, **kwargs):
    return NativeBedGraphReader(input_path, **kwargs)

  def _record_proto(self):
    return bedgraph_pb2.BedGraphRecord


class NativeBedGraphWriter(genomics_writer.GenomicsWriter):
  """Class for writing to native BedGraph files.

  Most users will want BedGraphWriter, which will write to either native
  BedGraph files or TFRecord files, based on the output filename's extension.
  """

  def __init__(self, output_path, header=None):
    """Initializer for NativeBedGraphWriter.

    Args:
      output_path: str. The path to which to write the BedGraph file.
    """
    super(NativeBedGraphWriter, self).__init__()
    self._writer = bedgraph_writer.BedGraphWriter.to_file(output_path)

  def write(self, proto):
    self._writer.write(proto)

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._writer.__exit__(exit_type, exit_value, exit_traceback)


class BedGraphWriter(genomics_writer.DispatchingGenomicsWriter):
  """Class for writing BedGraphRecord protos to BedGraph or TFRecord files."""

  def _native_writer(self, output_path):
    return NativeBedGraphWriter(output_path)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nucleus/protos/range.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='nucleus/protos/range.proto',
  package='nucleus.genomics.v1',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1anucleus/protos/range.proto\x12\x13nucleus.genomics.v1\";\n\x05Range\x12\x16\n\x0ereference_name\x18\x01 \x01(\t\x12\r\n\x05start\x18\x02 \x01(\x03\x12\x0b\n\x03\x65nd\x18\x03 \x01(\x03\x62\x06proto3')
)




_RANGE = _descriptor.Descriptor(
  name='Range',
  full_name='nucleus.genomics.v1.Range',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='reference_name', full_name='nucleus.genomics.v1.Range.reference_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start', full_name='nucleus.genomics.v1.Range.start', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='end', full_name='nucleus.genomics.v1.Range.end', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=51,
  serialized_end=110,
)

DESCRIPTOR.message_types_by_name['Range'] = _RANGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Range = _reflection.GeneratedProtocolMessageType('Range', (_message.Message,), {
  'DESCRIPTOR' : _RANGE,
  '__module__' : 'nucleus.protos.range_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.Range)
  })
_sym_db.RegisterMessage(Range)


# @@protoc_insertion_point(module_scope)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for working with reads."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from nucleus.protos import range_pb2
from nucleus.util.python import utils as utils_cpp


def read_overlaps_region(read, region):
  """Returns True if read overlaps read.

  This function is equivalent to calling:

    `ranges.ranges_overlap(region, read_range(read))`

  But is optimized for speed and memory performance in C++.

  Args:
    read: nucleus.genomics.v1.Read.
    region: nucleus.genomics.v1.Range.

  Returns:
    True if read and region overlap (i.e, have the same reference_name and their
    start/ends overlap at least one basepair).
  """
  return utils_cpp.read_overlaps_region(read, region)


def read_range(read):
  """Creates a Range proto from the alignment of Read.

  Args:
    read: nucleus.genomics.v1.Read. The read to calculate the range for.

  Returns:
    A nucleus.genomics.v1.Range for read.
  """
  range_pb = range_pb2.Range()
  utils_cpp.read_range(read, range_pb)
  return range_pb


def read_end(read):
  """Returns the read start + alignment length for Read read."""
  return read_range(read).end


def reservoir_sample(iterable, k, random=None):
  """Samples k elements with uniform probability from an iterable.

  Selects a subset of k elements from n input elements with uniform probability
  without needing to hold all n elements in memory at the same time. This
  implementation has max space complexity O(min(k, n)), i.e., we allocate up to
  min(k, n) elements to store the samples. This means that we only use ~n
  elements when n is smaller than k, which can be important when k is large. If
  n elements are added to this sampler, and n <= k, all n elements will be
  retained. If n > k, each added element will be retained with a uniform
  probability of k / n.

  The order of the k retained samples from our n elements is undefined. In
  particular that means that the elements in the returned list can occur in a
  different order than they appeared in the iterable.

  More details about reservoir sampling (and the specific algorithm used here
  called Algorithm R) can be found on wikipedia:

  https://en.wikipedia.org/wiki/Reservoir_sampling#Algorithm_R

  Args:
    iterable: Python iterable. The iterable to sample from.
    k: int. The number of elements to sample.
    random: A random number generator or None.

  Returns:
    A list containing the k sampled elements.

  Raises:
    ValueError: If k is negative.
  """
  if k < 0:
    raise ValueError('k must be nonnegative, but got {}'.format(k))
  if random is None:
    random = np.random
  sample = []
  for i, item in enumerate(iterable):
    if len(sample) < k:
      sample.append(item)
    else:
      j = random.randint(0, i + 1)
      if j < k:
        sample[j] = item
  return sample



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for reading FASTA files.

The FASTA format is described at
https://en.wikipedia.org/wiki/FASTA_format

API for reading:

```python
from nucleus.io import fasta
from nucleus.protos import range_pb2

with fasta.IndexedFastaReader(input_path) as reader:
  region = range_pb2.Range(reference_name='chrM', start=1, end=6)
  basepair_string = reader.query(region)
  print(basepair_string)
```

If `input_path` ends with '.gz', it is assumed to be compressed.  All FASTA
files are assumed to be indexed with the index file located at
`input_path + '.fai'`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

import six

from nucleus.io import gfile
from nucleus.io import genomics_reader
from nucleus.io.python import reference
from nucleus.protos import fasta_pb2
from nucleus.protos import reference_pb2
from nucleus.util import ranges

# TODO(thomaswc): Replace this with a real protocol buffer definition.
RefFastaHeader = collections.namedtuple(
    'RefFastaHeader', ['contigs'])


class FastaReader(genomics_reader.DispatchingGenomicsReader):
  """Class for reading (name, bases) tuples from FASTA files."""

  def _native_reader(self, input_path, **kwargs):
    fai_path = input_path + '.fai'
    if gfile.Exists(fai_path):
      return IndexedFastaReader(input_path, **kwargs)
    return UnindexedFastaReader(input_path, **kwargs)

  def _record_proto(self):
    return fasta_pb2.FastaRecord


class IndexedFastaReader(genomics_reader.GenomicsReader):
  """Class for reading from FASTA files containing a reference genome."""

  def __init__(self, input_path, keep_true_case=False, cache_size=None):
    """Initializes an IndexedFastaReader.

    Args:
      input_path: string. A path to a resource containing FASTA records.
      keep_true_case: bool. If False, casts all bases to uppercase before
        returning them.
      cache_size: integer. Number of bases to cache from previous queries.
        Defaults to 64K.  The cache can be disabled using cache_size=0.
    """
    super(IndexedFastaReader, self).__init__()

    options = fasta_pb2.FastaReaderOptions(keep_true_case=keep_true_case)

    fasta_path = input_path
    fai_path = fasta_path + '.fai'
    if cache_size is None:
      # Use the C++-defined default cache size.
      self._reader = reference.IndexedFastaReader.from_file(
          fasta_path, fai_path, options)
    else:
      self._reader = reference.IndexedFastaReader.from_file(
          fasta_path, fai_path, options, cache_size)

    # TODO(thomaswc): Define a RefFastaHeader proto, and use it instead of this.
    self.header = RefFastaHeader(contigs=self._reader.contigs)

  def iterate(self):
    """Returns an iterable of (name, bases) tuples contained in this file."""
    return self._reader.iterate()

  def query(self, region):
    """Returns the base pairs (as a string) in the given region."""
    return self._reader.bases(region)

  def is_valid(self, region):
    """Returns whether the region is contained in this FASTA file."""
    return self._reader.is_valid_interval(region)

  def contig(self, contig_name):
    """Returns a ContigInfo proto for contig_name."""
    return self._reader.contig(contig_name)

  @property
  def c_reader(self):
    """Returns the underlying C++ reader."""
    return self._reader

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._reader.__exit__(exit_type, exit_value, exit_traceback)


class UnindexedFastaReader(genomics_reader.GenomicsReader):
  """Class for reading from unindexed FASTA files."""

  def __init__(self, input_path):
    """Initializes an UnindexedFastaReader.

    Args:
      input_path: string. A path to a resource containing FASTA records.
    """
    super(UnindexedFastaReader, self).__init__()

    self._reader = reference.UnindexedFastaReader.from_file(input_path)

  def iterate(self):
    """Returns an iterable of (name, bases) tuples contained in this file."""
    return self._reader.iterate()

  def query(self, region):
    """Returns the base pairs (as a string) in the given region."""
    raise NotImplementedError('Can not query an unindexed FASTA file')

  def is_valid(self, region):
    """Returns whether the region is contained in this FASTA file."""
    return self._reader.is_valid_interval(region)

  def contig(self, contig_name):
    """Returns a ContigInfo proto for contig_name."""
    raise NotImplementedError('Contigs unknown for an unindexed FASTA file')

  @property
  def c_reader(self):
    """Returns the underlying C++ reader."""
    return self._reader

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._reader.__exit__(exit_type, exit_value, exit_traceback)


class InMemoryFastaReader(genomics_reader.GenomicsReader):
  """An `IndexedFastaReader` getting its bases from an in-memory data structure.

  An `InMemoryFastaReader` provides the same API as `IndexedFastaReader` but
  doesn't fetch its data from an on-disk FASTA file but rather fetches the bases
  from an in-memory cache containing (chromosome, start, bases) tuples.

  In particular, the `query(Range(chrom, start, end))` operation fetches bases
  from the tuple where `chrom` == chromosome, and then from the bases where the
  first base of bases starts at start. If start > 0, then the bases string is
  assumed to contain bases starting from that position in the region. For
  example, the record ('1', 10, 'ACGT') implies that
  `query(ranges.make_range('1', 11, 12))` will return the base 'C', as the 'A'
  base is at position 10. This makes it straightforward to cache a small region
  of a full chromosome without having to store the entire chromosome sequence in
  memory (potentially big!).
  """

  def __init__(self, chromosomes):
    """Initializes an InMemoryFastaReader using data from chromosomes.

    Args:
      chromosomes: list[tuple]. The chromosomes we are caching in memory as a
        list of tuples. Each tuple must be exactly three elements in length,
        containing (chromosome name [str], start [int], bases [str]).

    Raises:
      ValueError: If any of the chromosomes tuples are invalid.
    """
    super(InMemoryFastaReader, self).__init__()

    ref_seqs = []
    contigs = []
    for i, (contig_name, start, bases) in enumerate(chromosomes):
      if start < 0:
        raise ValueError('start={} must be >= for chromosome={}'.format(
            start, contig_name))
      if not bases:
        raise ValueError(
            'Bases must contain at least one base, but got "{}"'.format(bases))

      end = start + len(bases)
      ref_seqs.append(reference_pb2.ReferenceSequence(
          region=ranges.make_range(contig_name, start, end), bases=bases))
      contigs.append(
          reference_pb2.ContigInfo(
              name=contig_name, n_bases=end, pos_in_fasta=i))

    self._reader = reference.InMemoryFastaReader.create(contigs, ref_seqs)
    self.header = RefFastaHeader(contigs=self._reader.contigs)

  def iterate(self):
    """Returns an iterable of (name, bases) tuples contained in this file."""
    return self._reader.iterate()

  def query(self, region):
    """Returns the base pairs (as a string) in the given region."""
    return self._reader.bases(region)

  def is_valid(self, region):
    """Returns whether the region is contained in this FASTA file."""
    return self._reader.is_valid_interval(region)

  def contig(self, contig_name):
    """Returns a ContigInfo proto for contig_name."""
    return self._reader.contig(contig_name)

  @property
  def c_reader(self):
    """Returns the underlying C++ reader."""
    return self._reader

  def __str__(self):

    def _format_refseq(refseq):
      bases = refseq.bases
      if len(bases) >= 50:
        bases = bases[0:50] + '...'
      return 'Contig(chrom={} start={}, end={}, bases={})'.format(
          refseq.region.reference_name, refseq.region.start, refseq.region.end,
          bases)

    contigs_strs = [
        _format_refseq(refseq)
        for refseq in six.itervalues(self._reader.reference_sequences)
    ]
    return 'InMemoryFastaReader(contigs={})'.format(''.join(contigs_strs))

  __repr__ = __str__



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for working with sharded files.

A sharded file is a single conceptual file that is broken into a collection
of files to make parallelization easier.  A sharded file spec is like a
filename for a sharded file; the file spec "/some/path/prefix@200.txt"
says that the sharded file consists of 200 actual files that have names like
"/some/path/prefix-00000-of-00200.txt", "/some/path/prefix-00001-of-00200.txt",
etc.  This module contains functions for parsing, generating, detecting and
resolving sharded file specs.
"""

# Important: Please keep this module free of TensorFlow c++ extensions.
# This makes it easy to build pure python packages for training that work with
# CMLE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import re
import six

from nucleus.io import gfile

SHARD_SPEC_PATTERN = re.compile(R'((.*)\@(\d*[1-9]\d*)(?:\.(.+))?)')


class ShardError(Exception):
  """An I/O error."""


def parse_sharded_file_spec(spec):
  """Parse a sharded file specification.

  Args:
    spec: str. The sharded file specification. A sharded file spec is one like
      'gs://some/file@200.txt'. Here, '@200' specifies the number of shards.

  Returns:
    basename: str. The basename for the files.
    num_shards: int >= 0. The number of shards.
    suffix: str. The suffix if there is one, or '' if not.
  Raises:
    ShardError: If the spec is not a valid sharded specification.
  """
  m = SHARD_SPEC_PATTERN.match(spec)
  if not m:
    raise ShardError(('The file specification {0} is not a sharded file '
                      'specification because it did not match the regex '
                      '{1}').format(spec, SHARD_SPEC_PATTERN.pattern))

  # If there's a non-empty suffix, we need to prepend '.' so we get files like
  # foo@20.ext instead of foo@ext. The original C++ parser version has:
  # string ext = StrCat(suff.empty() ? "" : ".", suff);
  suffix = '.' + m.group(4) if m.group(4) else ''

  return m.group(2), int(m.group(3)), suffix


def _shard_width(num_shards):
  """Return the width of the shard matcher based on the number of shards."""
  return max(5, int(math.floor(math.log10(num_shards)) + 1))


def generate_sharded_filenames(spec):
  """Generate the list of filenames corresponding to the sharding path.

  Args:
    spec: str. Represents a filename with a sharding specification.
      e.g., 'gs://some/file@200.txt' represents a file sharded 200 ways.

  Returns:
    List of filenames.

  Raises:
    ShardError: If spec is not a valid sharded file specification.
  """
  basename, num_shards, suffix = parse_sharded_file_spec(spec)
  files = []
  width = _shard_width(num_shards)
  format_str = '{{0}}-{{1:0{0}}}-of-{{2:0{0}}}{{3}}'.format(width)
  for i in range(num_shards):
    files.append(format_str.format(basename, i, num_shards, suffix))

  return files


def glob_list_sharded_file_patterns(comma_separated_patterns, sep=','):
  """Generate list of filenames corresponding to `comma_separated_patterns`.

  Args:
    comma_separated_patterns: str. A pattern or a comma-separated list of
      patterns that represent file names.
    sep: char. Separator character.

  Returns:
    List of filenames, sorted and dedupped.
  """
  return sorted(set([
      f
      for pattern in comma_separated_patterns.split(sep)
      for f in gfile.Glob(normalize_to_sharded_file_pattern(pattern))
  ]))


def generate_sharded_file_pattern(basename, num_shards, suffix):
  """Generate a sharded file pattern.

  Args:
    basename: str. The basename for the files.
    num_shards: int. The number of shards.
    suffix: str. The suffix if there is one or ''.
  Returns:
    pattern:
  """
  width = _shard_width(num_shards)
  specifier = '?' * width
  format_str = '{{0}}-{{1}}-of-{{2:0{0}}}{{3}}'.format(width)
  return format_str.format(basename, specifier, num_shards, suffix)


def normalize_to_sharded_file_pattern(spec_or_pattern):
  """Take a sharding spec or sharding file pattern and return a sharded pattern.

  The input can be a sharding spec(e.g '/some/file@10') or a sharded file
  pattern (e.g. '/some/file-?????-of-00010)

  Args:
    spec_or_pattern: str. A sharded file specification or sharded file pattern.

  Returns:
    A sharded file pattern.
  """
  try:
    basename, num_shards, suffix = parse_sharded_file_spec(spec_or_pattern)
  except ShardError:
    return spec_or_pattern
  return generate_sharded_file_pattern(basename, num_shards, suffix)


def is_sharded_file_spec(spec):
  """Returns True if spec is a sharded file specification."""
  m = SHARD_SPEC_PATTERN.match(spec)
  return m is not None


# TODO(mdepristo): retire when GenerateShardedFilename is added to library.
def sharded_filename(spec, i):
  """Gets a path appropriate for writing the ith file of a sharded spec."""
  return generate_sharded_filenames(spec)[i]


# TODO(b/64046543): Improve the return value (instead of using tuple). It hurts
# readability when there are multiple input filespecs.
def resolve_filespecs(shard, *filespecs):
  """Transforms potentially sharded filespecs into their paths for single shard.

  This function takes a shard number and a varargs of potentially-sharded
  filespecs, and returns a list where the filespecs have been resolved into
  concrete file paths for a single shard.

  This function has a concept of a master filespec, which is used to constrain
  and check the validity of other filespecs. The first filespec is considered
  the master, and it cannot be None. For example, if master is not sharded, none
  of the other specs can be sharded, and vice versa. They must all also have a
  consistent sharding (e.g., master is @10, then all others must be @10).

  Note that filespecs (except the master) may be None or any other False value,
  which are returned as-is in the output list.

  Args:
    shard: int >= 0. Our shard number.
    *filespecs: list[str]. Contains all of the filespecs we want to resolve into
      shard-specific file paths.

  Returns:
    A list. The first element is the number of shards, which is an int >= 1 when
    filespecs contains sharded paths and 0 if none do. All subsequent
    returned values follow the shard-specific paths for each filespec, in order.

  Raises:
    ValueError: if any filespecs are inconsistent.
  """
  if not filespecs:
    raise ValueError('filespecs must have at least one element.')

  master = filespecs[0]
  master_is_sharded = is_sharded_file_spec(master)

  master_num_shards = 0
  if master_is_sharded:
    _, master_num_shards, _ = parse_sharded_file_spec(master)
    if shard >= master_num_shards or shard < 0:
      raise ValueError('Invalid shard={} value with master={} sharding'.format(
          shard, master))
  elif shard > 0:
    raise ValueError('Output is not sharded but shard > 0: {}'.format(shard))

  def resolve_one(filespec):
    """Resolves a single filespec into a concrete filepath."""
    if not filespec:
      return filespec

    is_sharded = is_sharded_file_spec(filespec)
    if master_is_sharded != is_sharded:
      raise ValueError('Master={} and {} have inconsistent sharding'.format(
          master, filespec))

    if not is_sharded:  # Not sharded => filespec is the actual filename.
      return filespec

    _, filespec_num_shards, _ = parse_sharded_file_spec(filespec)
    if filespec_num_shards != master_num_shards:
      raise ValueError('Master={} and {} have inconsistent sharding'.format(
          master, filespec))
    return sharded_filename(filespec, shard)

  return [master_num_shards] + [resolve_one(spec) for spec in filespecs]


def maybe_generate_sharded_filenames(filespec):
  """Potentially expands sharded filespec into a list of paths.

  This function takes in a potentially sharded filespec and expands it into a
  list containing the full set of corresponding concrete sharded file paths. If
  the input filespec is not sharded then a list containing just that file path
  is returned. This function is useful, for example, when the input to a binary
  can either be sharded or not.

  Args:
    filespec: String. A potentially sharded filespec to expand.

  Returns:
    A list of file paths.

  Raises:
    TypeError: if filespec is not in valid string_types.
  """
  if not isinstance(filespec, six.string_types):
    raise TypeError('Invalid filespec: %s' % filespec)
  if is_sharded_file_spec(filespec):
    return generate_sharded_filenames(filespec)
  else:
    return [filespec]
